# encoding: utf-8

"""Unit test suite for the cr.cube.crunch_cube module."""

import numpy as np
import pytest

from unittest import TestCase

from cr.cube.crunch_cube import (
    _BaseMeasure,
    CrunchCube,
    _MeanMeasure,
    _Measures,
    _UnweightedCountMeasure,
    _WeightedCountMeasure,
)
from cr.cube.cube_slice import CubeSlice
from cr.cube.dimension import AllDimensions, _ApparentDimensions, Dimension
from cr.cube.enums import DIMENSION_TYPE as DT

from ...unitutil import (
    class_mock,
    function_mock,
    instance_mock,
    method_mock,
    Mock,
    patch,
    property_mock,
)


class DescribeCrunchCube(object):
    @pytest.mark.parametrize(
        "slices_can_compare_pairwise, cube_can_compare_pairwise",
        (
            ([True, True, True], True),
            ([False, True, True], False),
            ([True, True, False], False),
            ([False, False, False], False),
        ),
    )
    def it_can_compare_pariwise(
        self,
        request,
        slices_can_compare_pairwise,
        cube_can_compare_pairwise,
        slices_prop_,
    ):
        slices = [
            instance_mock(request, CubeSlice, can_compare_pairwise=can_compare_pairwise)
            for can_compare_pairwise in slices_can_compare_pairwise
        ]
        slices_prop_.return_value = slices
        cube = CrunchCube(None)
        assert cube.can_compare_pairwise is cube_can_compare_pairwise

    def it_provides_a_default_repr(self):
        cube = CrunchCube(None)
        repr_ = repr(cube)
        assert repr_.startswith("<cr.cube.crunch_cube.CrunchCube object at 0x")

    @pytest.mark.parametrize("weighted, expected_value", ((False, 42), (True, 48.732)))
    def it_knows_its_row_count(
        self, weighted, expected_value, _measures_prop_, measures_
    ):
        _measures_prop_.return_value = measures_
        measures_.unweighted_n = 42
        measures_.weighted_n = 48.732
        cube = CrunchCube(None)

        count = cube.count(weighted)

        assert count == expected_value

    def it_provides_access_to_its_dimensions(
        self, _all_dimensions_prop_, all_dimensions_, apparent_dimensions_
    ):
        _all_dimensions_prop_.return_value = all_dimensions_
        all_dimensions_.apparent_dimensions = apparent_dimensions_
        cube = CrunchCube(None)

        dimensions = cube.dimensions

        assert dimensions is apparent_dimensions_

    def it_knows_the_types_of_its_dimension(self, request, dimensions_prop_):
        dimensions_prop_.return_value = tuple(
            instance_mock(request, Dimension, name="dim-%d" % idx, dimension_type=dt)
            for idx, dt in enumerate((DT.CAT, DT.CA_SUBVAR, DT.MR, DT.MR_CAT))
        )
        cube = CrunchCube(None)

        dim_types = cube.dim_types

        assert dim_types == (DT.CAT, DT.CA_SUBVAR, DT.MR, DT.MR_CAT)

    @pytest.mark.parametrize(
        "has_mean_measure, expected_value", ((True, True), (False, False))
    )
    def it_knows_when_it_contains_means_data(
        self,
        has_mean_measure,
        expected_value,
        _measures_prop_,
        measures_,
        mean_measure_,
    ):
        means = mean_measure_ if has_mean_measure else None
        _measures_prop_.return_value = measures_
        measures_.means = means
        cube = CrunchCube(None)

        has_means = cube.has_means

        assert has_means is expected_value

    @pytest.mark.parametrize(
        "mr_dim_indices, expected_value",
        ((None, False), (0, True), (3, True), ([2, 5], True)),
    )
    def it_knows_when_it_has_an_mr_dimension(
        self, mr_dim_indices, expected_value, mr_dim_ind_prop_
    ):
        mr_dim_ind_prop_.return_value = mr_dim_indices
        cube = CrunchCube(None)

        has_mr = cube.has_mr

        assert has_mr is expected_value

    def it_has_a_deprecated_index_method_that_forwards_to_Index_data(self, request):
        Index_ = class_mock(request, "cr.cube.crunch_cube.Index")
        index_ = Index_.data.return_value
        warn_ = function_mock(request, "cr.cube.crunch_cube.warnings.warn")
        cube = CrunchCube(None)

        index = cube.index()

        warn_.assert_called_once_with(
            "CrunchCube.index() is deprecated. Use CubeSlice.index_table().",
            DeprecationWarning,
        )
        Index_.data.assert_called_once_with(cube, True, False)
        assert index is index_

    @pytest.mark.parametrize(
        "is_weighted, expected_value", ((True, True), (False, False))
    )
    def it_knows_when_it_contains_weighted_data(
        self, is_weighted, expected_value, _measures_prop_, measures_
    ):
        _measures_prop_.return_value = measures_
        measures_.is_weighted = is_weighted
        cube = CrunchCube(None)

        is_weighted = cube.is_weighted

        assert is_weighted is expected_value

    def it_knows_its_missing_count(self, _measures_prop_, measures_):
        _measures_prop_.return_value = measures_
        measures_.missing_count = 36
        cube = CrunchCube(None)

        missing = cube.missing

        assert missing == 36

    def it_knows_its_population_fraction(self, _measures_prop_, measures_):
        _measures_prop_.return_value = measures_
        measures_.population_fraction = 0.42
        cube = CrunchCube(None)

        population_fraction = cube.population_fraction

        assert population_fraction == 0.42

    @pytest.mark.parametrize(
        "dimension_types, axis_cases",
        (
            (
                (DT.CAT, DT.CAT),
                ((0, (0,)), (1, (1,)), (None, (0, 1)), ((0, 1), (0, 1))),
            ),
            (
                (DT.CAT, DT.CAT, DT.CAT),
                ((0, (0,)), (1, (1,)), (2, (2,)), (None, (1, 2)), ((1, 2), (1, 2))),
            ),
            ((DT.MR,), ((0, (1,)), (None, (1,)))),
            ((DT.CAT, DT.MR), ((0, (0,)), (1, (2,)), (None, (0, 2)), ((0, 1), (0, 2)))),
            ((DT.MR, DT.CAT), ((0, (1,)), (1, (2,)), (None, (1, 2)), ((0, 1), (1, 2)))),
            ((DT.MR, DT.MR), ((0, (1,)), (1, (3,)), (None, (1, 3)), ((0, 1), (1, 3)))),
            (
                (DT.CAT, DT.MR, DT.MR),
                ((0, (0,)), (1, (2,)), (2, (4,)), (None, (2, 4)), ((1, 2), (2, 4))),
            ),
            (
                (DT.MR, DT.CAT, DT.MR),
                ((0, (1,)), (1, (2,)), (2, (4,)), (None, (2, 4)), ((1, 2), (2, 4))),
            ),
            (
                (DT.MR, DT.MR, DT.CAT),
                ((0, (1,)), (1, (3,)), (2, (4,)), (None, (3, 4)), ((1, 2), (3, 4))),
            ),
            ((DT.CA, DT.CAT), ((1, (1,)),)),
            (
                (DT.CA, DT.CAT, DT.CAT),
                ((1, (1,)), (2, (2,)), (None, (1, 2)), ((1, 2), (1, 2))),
            ),
            ((DT.CAT, DT.CA, DT.CAT), ((2, (2,)),)),
            (
                (DT.CA, DT.CAT, DT.MR),
                ((1, (1,)), (2, (3,)), (None, (1, 3)), ((1, 2), (1, 3))),
            ),
            ((DT.MR, DT.CA, DT.CAT), ((0, (1,)), (2, (3,)))),
        ),
    )
    def it_can_adjust_an_axis_to_help(
        self, request, dimension_types, axis_cases, dimensions_prop_
    ):
        dimensions_prop_.return_value = tuple(
            instance_mock(request, Dimension, dimension_type=dimension_type)
            for dimension_type in dimension_types
        )
        cube = CrunchCube(None)

        for axis, expected_value in axis_cases:
            adjusted_axis = cube._adjust_axis(axis)
            assert adjusted_axis == expected_value

    def but_it_raises_on_disallowed_adjustment(self, _is_axis_allowed_):
        _is_axis_allowed_.return_value = False
        axis = 42
        cube = CrunchCube(None)

        with pytest.raises(ValueError):
            cube._adjust_axis(axis)

        _is_axis_allowed_.assert_called_once_with(cube, axis)

    def it_provides_its_AllDimensions_collection_to_help(
        self, AllDimensions_, all_dimensions_
    ):
        cube_response = {"result": {"dimensions": [{"d": 1}, {"d": 2}]}}
        AllDimensions_.return_value = all_dimensions_
        cube = CrunchCube(cube_response)

        all_dimensions = cube._all_dimensions

        AllDimensions_.assert_called_once_with([{"d": 1}, {"d": 2}])
        assert all_dimensions is all_dimensions_

    def it_provides_access_to_the_cube_dict_to_help(self):
        cube = CrunchCube({"cube": "dict"})
        cube_dict = cube._cube_dict
        assert cube_dict == {"cube": "dict"}

    def but_it_first_parses_a_JSON_cube_response(self):
        cube = CrunchCube('{"cubic": "dictum"}')
        cube_dict = cube._cube_dict
        assert cube_dict == {"cubic": "dictum"}

    def and_it_raises_on_invalid_cube_response_type(self):
        cube = CrunchCube(42)
        with pytest.raises(TypeError) as e:
            cube._cube_dict
        assert str(e.value).startswith("Unsupported type")

    @pytest.mark.parametrize(
        "dimension_types, axis_cases",
        (
            ((DT.CA, DT.CAT), ((0, False), (1, True), (None, False))),
            (
                (DT.CA, DT.CAT, DT.CAT),
                ((0, False), (1, True), (2, True), (None, True), ((1, 2), True)),
            ),
            (
                (DT.CAT, DT.CA, DT.CAT),
                ((0, True), (1, False), (2, True), (None, False), ((1, 2), False)),
            ),
            (
                (DT.MR, DT.CA, DT.CAT),
                ((0, True), (1, False), (2, True), (None, False), ((1, 2), False)),
            ),
            (
                (DT.CA, DT.CAT, DT.MR),
                ((0, False), (1, True), (2, True), (None, True), ((1, 2), True)),
            ),
            ((DT.CAT,), ((0, True), (1, True), (None, True))),
            ((DT.CAT, DT.CAT), ((0, True), (1, True), (None, True), ((0, 1), True))),
            (
                (DT.CAT, DT.MR, DT.MR),
                ((0, True), (1, True), (2, True), (None, True), ((1, 2), True)),
            ),
            (
                (DT.MR, DT.CAT, DT.MR),
                ((0, True), (1, True), (2, True), (None, True), ((1, 2), True)),
            ),
            (
                (DT.MR, DT.MR, DT.CAT),
                ((0, True), (1, True), (2, True), (None, True), ((1, 2), True)),
            ),
        ),
    )
    def it_knows_whether_an_axis_is_marginable_to_help(
        self, request, dimension_types, axis_cases, dimensions_prop_
    ):
        dimensions_prop_.return_value = tuple(
            instance_mock(request, Dimension, dimension_type=dimension_type)
            for dimension_type in dimension_types
        )
        cube = CrunchCube(None)

        for axis, expected_value in axis_cases:
            axis_is_marginable = cube._is_axis_allowed(axis)
            assert axis_is_marginable is expected_value

    @pytest.mark.parametrize(
        "weighted, is_weighted, expected_type",
        (
            (False, False, "_UnweightedCountMeasure"),
            (False, True, "_UnweightedCountMeasure"),
            (True, False, "_UnweightedCountMeasure"),
            (True, True, "_WeightedCountMeasure"),
        ),
    )
    def it_selects_the_best_match_counts_measure_to_help(
        self,
        weighted,
        is_weighted,
        expected_type,
        _measures_prop_,
        measures_,
        weighted_count_measure_,
        unweighted_count_measure_,
    ):
        measures_.weighted_counts = (
            weighted_count_measure_ if is_weighted else unweighted_count_measure_
        )
        measures_.unweighted_counts = unweighted_count_measure_
        expected_measure = {
            "_UnweightedCountMeasure": unweighted_count_measure_,
            "_WeightedCountMeasure": weighted_count_measure_,
        }[expected_type]
        _measures_prop_.return_value = measures_
        cube = CrunchCube(None)

        measure = cube._counts(weighted)

        assert measure is expected_measure

    @pytest.mark.parametrize(
        "has_means, weighted, is_weighted, expected_type",
        (
            (True, False, False, "_MeanMeasure"),
            (True, False, True, "_MeanMeasure"),
            (True, True, False, "_MeanMeasure"),
            (True, True, True, "_MeanMeasure"),
            (False, False, False, "_UnweightedCountMeasure"),
            (False, False, True, "_UnweightedCountMeasure"),
            (False, True, False, "_UnweightedCountMeasure"),
            (False, True, True, "_WeightedCountMeasure"),
        ),
    )
    def it_selects_the_best_match_measure_to_help(
        self,
        has_means,
        weighted,
        is_weighted,
        expected_type,
        _measures_prop_,
        measures_,
        unweighted_count_measure_,
        weighted_count_measure_,
        mean_measure_,
    ):
        measures_.means = mean_measure_ if has_means else None
        measures_.weighted_counts = (
            weighted_count_measure_ if is_weighted else unweighted_count_measure_
        )
        measures_.unweighted_counts = unweighted_count_measure_
        expected_measure = {
            "_MeanMeasure": mean_measure_,
            "_UnweightedCountMeasure": unweighted_count_measure_,
            "_WeightedCountMeasure": weighted_count_measure_,
        }[expected_type]
        _measures_prop_.return_value = measures_
        cube = CrunchCube(None)

        measure = cube._measure(weighted)

        assert measure is expected_measure

    def it_provides_access_to_its_Measures_object_to_help(
        self, _all_dimensions_prop_, all_dimensions_, _Measures_, measures_
    ):
        cube_dict = {"cube": "dict"}
        _all_dimensions_prop_.return_value = all_dimensions_
        _Measures_.return_value = measures_
        cube = CrunchCube(cube_dict)

        measures = cube._measures

        _Measures_.assert_called_once_with(cube_dict, all_dimensions_)
        assert measures is measures_

    # fixture components ---------------------------------------------

    @pytest.fixture
    def AllDimensions_(self, request):
        return class_mock(request, "cr.cube.crunch_cube.AllDimensions")

    @pytest.fixture
    def slices_prop_(self, request):
        return property_mock(request, CrunchCube, "slices")

    @pytest.fixture
    def all_dimensions_(self, request):
        return instance_mock(request, AllDimensions)

    @pytest.fixture
    def _all_dimensions_prop_(self, request):
        return property_mock(request, CrunchCube, "_all_dimensions")

    @pytest.fixture
    def apparent_dimensions_(self, request):
        return instance_mock(request, _ApparentDimensions)

    @pytest.fixture
    def dimensions_prop_(self, request):
        return property_mock(request, CrunchCube, "dimensions")

    @pytest.fixture
    def _is_axis_allowed_(self, request):
        return method_mock(request, CrunchCube, "_is_axis_allowed")

    @pytest.fixture
    def mean_measure_(self, request):
        return instance_mock(request, _MeanMeasure)

    @pytest.fixture
    def _Measures_(self, request):
        return class_mock(request, "cr.cube.crunch_cube._Measures")

    @pytest.fixture
    def measures_(self, request):
        return instance_mock(request, _Measures)

    @pytest.fixture
    def _measures_prop_(self, request):
        return property_mock(request, CrunchCube, "_measures")

    @pytest.fixture
    def mr_dim_ind_prop_(self, request):
        return property_mock(request, CrunchCube, "mr_dim_ind")

    @pytest.fixture
    def unweighted_count_measure_(self, request):
        return instance_mock(request, _UnweightedCountMeasure)

    @pytest.fixture
    def weighted_count_measure_(self, request):
        return instance_mock(request, _WeightedCountMeasure)


class Describe_Measures(object):
    @pytest.mark.parametrize(
        "cube_dict, expected_value",
        (
            ({"query": {"weight": "https://x"}}, True),
            ({"weight_var": "weight"}, True),
            ({"weight_url": "https://y"}, True),
            (
                {
                    "result": {
                        "counts": [1, 2, 3],
                        "measures": {"count": {"data": [2, 4, 6]}},
                    }
                },
                True,
            ),
            (
                {
                    "result": {
                        "counts": [1, 2, 3],
                        "measures": {"count": {"data": [1, 2, 3]}},
                    }
                },
                False,
            ),
        ),
    )
    def it_knows_when_a_measure_is_weighted(self, cube_dict, expected_value):
        measures = _Measures(cube_dict, None)

        is_weighted = measures.is_weighted

        assert is_weighted is expected_value

    def it_provides_access_to_the_means_measure(
        self, _MeanMeasure_, mean_measure_, all_dimensions_
    ):
        cube_dict = {"result": {"measures": {"mean": {}}}}
        _MeanMeasure_.return_value = mean_measure_
        measures = _Measures(cube_dict, all_dimensions_)

        means = measures.means

        _MeanMeasure_.assert_called_once_with(cube_dict, all_dimensions_)
        assert means is mean_measure_

    def but_only_when_the_cube_has_a_mean_measure(self):
        measures = _Measures({"result": {"measures": {}}}, None)
        means = measures.means
        assert means is None

    @pytest.mark.parametrize(
        "cube_dict, has_means, expected_value",
        (
            ({}, True, 37),
            ({"result": {"missing": 42}}, False, 42),
            ({"result": {}}, False, 0),
        ),
    )
    def it_knows_the_missing_count(
        self, cube_dict, has_means, expected_value, means_prop_, mean_measure_
    ):
        mean_measure_.missing_count = expected_value if has_means else None
        means = mean_measure_ if has_means else None
        means_prop_.return_value = means
        measures = _Measures(cube_dict, None)

        missing_count = measures.missing_count

        assert missing_count == expected_value

    @pytest.mark.parametrize(
        "cube_dict, expected_value",
        (
            ({"result": {}}, 1.0),
            (
                {
                    "result": {
                        "filtered": {"weighted_n": 21},
                        "unfiltered": {"weighted_n": 42},
                    }
                },
                0.5,
            ),
            (
                {
                    "result": {
                        "filtered": {"weighted_n": 0},
                        "unfiltered": {"weighted_n": 0},
                    }
                },
                np.nan,
            ),
            ({"result": {"filtered": {"weighted_n": 43}}}, 1.0),
            ({"result": {"unfiltered": {"weighted_n": 44}}}, 1.0),
        ),
    )
    def it_knows_the_population_fraction(self, cube_dict, expected_value):
        measures = _Measures(cube_dict, None)

        population_fraction = measures.population_fraction

        # ---works for np.nan, which doesn't equal itself---
        assert population_fraction in (expected_value,)

    def it_provides_access_to_the_unweighted_count_measure(
        self, _UnweightedCountMeasure_, unweighted_count_measure_, all_dimensions_
    ):
        cube_dict = {"cube": "dict"}
        _UnweightedCountMeasure_.return_value = unweighted_count_measure_
        measures = _Measures(cube_dict, all_dimensions_)

        unweighted_counts = measures.unweighted_counts

        _UnweightedCountMeasure_.assert_called_once_with(cube_dict, all_dimensions_)
        assert unweighted_counts is unweighted_count_measure_

    def it_knows_the_unweighted_n(self):
        cube_dict = {"result": {"n": 121}}
        measures = _Measures(cube_dict, None)

        unweighted_n = measures.unweighted_n

        assert unweighted_n == 121

    def it_provides_access_to_the_weighted_count_measure(
        self,
        is_weighted_prop_,
        _WeightedCountMeasure_,
        weighted_count_measure_,
        all_dimensions_,
    ):
        cube_dict = {"cube": "dict"}
        is_weighted_prop_.return_value = True
        _WeightedCountMeasure_.return_value = weighted_count_measure_
        measures = _Measures(cube_dict, all_dimensions_)

        weighted_counts = measures.weighted_counts

        _WeightedCountMeasure_.assert_called_once_with(cube_dict, all_dimensions_)
        assert weighted_counts is weighted_count_measure_

    def but_it_returns_unweighted_count_measure_when_cube_is_not_weighted(
        self, is_weighted_prop_, unweighted_counts_prop_, unweighted_count_measure_
    ):
        is_weighted_prop_.return_value = False
        unweighted_counts_prop_.return_value = unweighted_count_measure_
        measures = _Measures(None, None)

        weighted_counts = measures.weighted_counts

        assert weighted_counts is unweighted_count_measure_

    @pytest.mark.parametrize(
        " cube_dict, is_weighted, expected_value",
        (
            ({}, False, 24.0),
            ({"result": {"measures": {"count": {"data": [7, 9]}}}}, True, 16.0),
        ),
    )
    def it_knows_the_weighted_n(
        self,
        cube_dict,
        is_weighted,
        expected_value,
        unweighted_n_prop_,
        is_weighted_prop_,
    ):
        is_weighted_prop_.return_value = is_weighted
        unweighted_n_prop_.return_value = 24
        measures = _Measures(cube_dict, None)

        weighted_n = measures.weighted_n

        assert weighted_n == expected_value

    # fixture components ---------------------------------------------

    @pytest.fixture
    def all_dimensions_(self, request):
        return instance_mock(request, AllDimensions)

    @pytest.fixture
    def is_weighted_prop_(self, request):
        return property_mock(request, _Measures, "is_weighted")

    @pytest.fixture
    def _MeanMeasure_(self, request):
        return class_mock(request, "cr.cube.crunch_cube._MeanMeasure")

    @pytest.fixture
    def mean_measure_(self, request):
        return instance_mock(request, _MeanMeasure)

    @pytest.fixture
    def means_prop_(self, request):
        return property_mock(request, _Measures, "means")

    @pytest.fixture
    def _UnweightedCountMeasure_(self, request):
        return class_mock(request, "cr.cube.crunch_cube._UnweightedCountMeasure")

    @pytest.fixture
    def unweighted_count_measure_(self, request):
        return instance_mock(request, _UnweightedCountMeasure)

    @pytest.fixture
    def unweighted_counts_prop_(self, request):
        return property_mock(request, _Measures, "unweighted_counts")

    @pytest.fixture
    def unweighted_n_prop_(self, request):
        return property_mock(request, _Measures, "unweighted_n")

    @pytest.fixture
    def _WeightedCountMeasure_(self, request):
        return class_mock(request, "cr.cube.crunch_cube._WeightedCountMeasure")

    @pytest.fixture
    def weighted_count_measure_(self, request):
        return instance_mock(request, _WeightedCountMeasure)


class Describe_BaseMeasure(object):
    def it_provides_access_to_the_raw_cube_array(
        self, _flat_values_prop_, all_dimensions_
    ):
        _flat_values_prop_.return_value = (
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
            2,
            4,
            6,
            8,
            0,
            1,
            3,
            5,
            7,
        )
        all_dimensions_.shape = (2, 3, 3)
        base_measure = _BaseMeasure(None, all_dimensions_)

        raw_cube_array = base_measure.raw_cube_array

        np.testing.assert_array_equal(
            raw_cube_array,
            np.array(
                [[[1, 2, 3], [4, 5, 6], [7, 8, 9]], [[2, 4, 6], [8, 0, 1], [3, 5, 7]]]
            ),
        )
        assert raw_cube_array.flags.writeable is False

    def it_requires_each_subclass_to_implement_flat_values(self):
        base_measure = _BaseMeasure(None, None)

        with pytest.raises(NotImplementedError) as pt_exc_info:
            base_measure._flat_values

        exception = pt_exc_info.value
        assert str(exception) == "must be implemented by each subclass"

    # fixture components ---------------------------------------------

    @pytest.fixture
    def all_dimensions_(self, request):
        return instance_mock(request, AllDimensions)

    @pytest.fixture
    def _flat_values_prop_(self, request):
        return property_mock(request, _BaseMeasure, "_flat_values")


class Describe_MeanMeasure(object):
    @pytest.mark.parametrize(
        "cube_dict, expected_value",
        (
            ({"result": {"measures": {"mean": {}}}}, 0),
            ({"result": {"measures": {"mean": {"n_missing": 42}}}}, 42),
        ),
    )
    def it_knows_the_missing_count(self, cube_dict, expected_value):
        mean_measure = _MeanMeasure(cube_dict, None)

        missing_count = mean_measure.missing_count

        assert missing_count == expected_value

    def it_parses_the_flat_values_to_help(self):
        cube_dict = {"result": {"measures": {"mean": {"data": [1, 2, {"?": -1}, 4]}}}}
        mean_measure = _MeanMeasure(cube_dict, None)

        flat_values = mean_measure._flat_values

        assert flat_values == (1, 2, np.nan, 4)


class Describe_UnweightedCountMeasure(object):
    def it_parses_the_flat_values_to_help(self):
        cube_dict = {"result": {"counts": [1, 2, 3, 4]}}
        unweighted_count_measure = _UnweightedCountMeasure(cube_dict, None)

        flat_values = unweighted_count_measure._flat_values

        assert flat_values == (1, 2, 3, 4)


class Describe_WeightedCountMeasure(object):
    def it_parses_the_flat_values_to_help(self):
        cube_dict = {"result": {"measures": {"count": {"data": [1.1, 2.2, 3.3, 4.4]}}}}
        weighted_count_measure = _WeightedCountMeasure(cube_dict, None)

        flat_values = weighted_count_measure._flat_values

        assert flat_values == (1.1, 2.2, 3.3, 4.4)


# pylint: disable=invalid-name, no-self-use, protected-access
@patch("cr.cube.crunch_cube.CrunchCube.get_slices", lambda x: None)
class TestCrunchCube(TestCase):
    """Test class for the CrunchCube unit tests.

    This class also tests the functionality of private methods,
    not just the API ones.
    """

    def test_calculate_constraints_sum_axis_0(self):
        prop_table = np.array([[0.32, 0.21, 0.45], [0.12, 0.67, 0.73]])
        prop_cols_margin = np.array([0.2, 0.3, 0.5])
        axis = 0
        expected = np.dot((prop_table * (1 - prop_table)), prop_cols_margin)
        actual = CrunchCube._calculate_constraints_sum(
            prop_table, prop_cols_margin, axis
        )
        np.testing.assert_array_equal(actual, expected)

    def test_calculate_constraints_sum_axis_1(self):
        prop_table = np.array([[0.32, 0.21, 0.45], [0.12, 0.67, 0.73]])
        prop_rows_margin = np.array([0.34, 0.66])
        axis = 1
        expected = np.dot(prop_rows_margin, (prop_table * (1 - prop_table)))
        actual = CrunchCube._calculate_constraints_sum(
            prop_table, prop_rows_margin, axis
        )
        np.testing.assert_array_equal(actual, expected)

    def test_calculate_constraints_sum_raises_value_error_for_bad_axis(self):
        with self.assertRaises(ValueError):
            CrunchCube._calculate_constraints_sum(Mock(), Mock(), 2)

    @patch("cr.cube.crunch_cube.CrunchCube.dimensions", None)
    def test_name_with_no_dimensions(self):
        fake_cube = {}
        cube = CrunchCube(fake_cube)
        expected = None
        actual = cube.name
        self.assertEqual(actual, expected)

    @patch("cr.cube.crunch_cube.CrunchCube.dimensions")
    def test_name_with_one_dimension(self, mock_dims):
        fake_cube = {}
        cube = CrunchCube(fake_cube)
        mock_dims[0].name = "test"
        expected = "test"
        actual = cube.name
        self.assertEqual(actual, expected)

    @patch("cr.cube.crunch_cube.CrunchCube.dimensions", None)
    def test_description_with_no_dimensions(self):
        fake_cube = {}
        cube = CrunchCube(fake_cube)
        expected = None
        actual = cube.name
        self.assertEqual(actual, expected)

    @patch("cr.cube.crunch_cube.CrunchCube.dimensions")
    def test_description_with_one_dimension(self, mock_dims):
        fake_cube = {}
        cube = CrunchCube(fake_cube)
        mock_dims[0].description = "test"
        expected = "test"
        actual = cube.description
        self.assertEqual(actual, expected)

    def test_fix_valid_indices_subsequent(self):
        initial_indices = [[1, 2, 3]]
        insertion_index = 2
        expected = [[1, 2, 3, 4]]
        dimension = 0
        actual = CrunchCube._fix_valid_indices(
            initial_indices, insertion_index, dimension
        )
        self.assertEqual(actual, expected)

    def test_fix_valid_indices_with_gap(self):
        initial_indices = [[0, 1, 2, 5, 6]]
        insertion_index = 2
        expected = [[0, 1, 2, 3, 6, 7]]
        dimension = 0
        actual = CrunchCube._fix_valid_indices(
            initial_indices, insertion_index, dimension
        )
        self.assertEqual(actual, expected)

    def test_fix_valid_indices_zero_position(self):
        initial_indices = [[0, 1, 2, 5, 6]]
        insertion_index = -1
        expected = [[0, 1, 2, 3, 6, 7]]
        dimension = 0
        actual = CrunchCube._fix_valid_indices(
            initial_indices, insertion_index, dimension
        )
        self.assertEqual(actual, expected)

    @patch("cr.cube.crunch_cube.CrunchCube.dimensions", [])
    def test_does_not_have_description(self):
        expected = None
        actual = CrunchCube(None).description
        self.assertEqual(actual, expected)

    @patch("cr.cube.crunch_cube.CrunchCube.dimensions")
    def test_has_description(self, mock_dims):
        dims = [Mock(), Mock()]
        mock_dims.__get__ = Mock(return_value=dims)
        expected = dims[0].description
        actual = CrunchCube(None).description
        self.assertEqual(actual, expected)

    def test_test_filter_annotation(self):
        mock_cube = {"filter_names": Mock()}
        expected = mock_cube["filter_names"]
        actual = CrunchCube(mock_cube).filter_annotation
        self.assertEqual(actual, expected)

    def test_margin_pruned_indices_without_insertions(self):
        table = np.array([0, 1, 0, 2, 3, 4])
        expected = np.array([True, False, True, False, False, False])
        actual = CrunchCube._margin_pruned_indices(table, [], 0)
        np.testing.assert_array_equal(actual, expected)

    def test_margin_pruned_indices_without_insertions_with_nans(self):
        table = np.array([0, 1, 0, 2, 3, np.nan])
        expected = np.array([True, False, True, False, False, True])
        actual = CrunchCube._margin_pruned_indices(table, [], 0)
        np.testing.assert_array_equal(actual, expected)

    def test_margin_pruned_indices_with_insertions(self):
        table = np.array([0, 1, 0, 2, 3, 4])
        insertions = [0, 1]
        expected = np.array([False, False, True, False, False, False])

        actual = CrunchCube._margin_pruned_indices(table, insertions, 0)
        np.testing.assert_array_equal(actual, expected)

    def test_margin_pruned_indices_with_insertions_and_nans(self):
        table = np.array([0, 1, 0, 2, 3, np.nan])
        insertions = [0, 1]
        expected = np.array([False, False, True, False, False, True])

        actual = CrunchCube._margin_pruned_indices(table, insertions, 0)
        np.testing.assert_array_equal(actual, expected)

    @patch("numpy.array")
    @patch("cr.cube.crunch_cube.CrunchCube.inserted_hs_indices")
    @patch("cr.cube.crunch_cube.CrunchCube.ndim", 1)
    def test_inserted_inds(self, mock_inserted_hs_indices, mock_np_array):
        expected = Mock()
        mock_np_array.return_value = expected

        cc = CrunchCube(None)

        # Assert indices are not fetched without trasforms
        actual = cc._inserted_dim_inds(None, 0)
        # Assert np.array called with empty list as argument
        mock_np_array.assert_called_once_with([])
        assert actual == expected

        # Assert indices are fetch with transforms
        actual = cc._inserted_dim_inds([1], 0)
        mock_inserted_hs_indices.assert_not_called()
        actual = cc._inserted_dim_inds([0], 0)
        mock_inserted_hs_indices.assert_called_once()



# encoding: utf-8

"""Utilities for cr.cube tests."""

from __future__ import absolute_import, division, print_function, unicode_literals

import os


def load_expectation(expectation_file_name, strip=True):  # pragma: no cover
    """Return (unicode) str containing text in *expectation_file_name*.

    Expectation file path is rooted at tests/expectations.
    """
    thisdir = os.path.dirname(__file__)
    expectation_file_path = os.path.abspath(
        os.path.join(thisdir, "expectations", "%s.txt" % expectation_file_name)
    )
    with open(expectation_file_path, "rb") as f:
        expectation_bytes = f.read()
    if strip:
        return expectation_bytes.decode("utf-8").strip()
    return expectation_bytes.decode("utf-8")


def load_python_expression(expression_file_name):
    """Return a Python object (list, dict) formed by parsing `expression_file_name`.

    Expectation file path is rooted at tests/expectations.
    """
    thisdir = os.path.dirname(__file__)
    expression_file_path = os.path.abspath(
        os.path.join(thisdir, "expectations", "%s.py" % expression_file_name)
    )
    with open(expression_file_path) as f:
        expression_bytes = f.read()
    return eval(expression_bytes)



# encoding: utf-8

"""Unit test suite for `cr.cube.matrix` module."""

from __future__ import absolute_import, division, print_function, unicode_literals

import sys

import numpy as np
import pytest

from cr.cube.cube import Cube
from cr.cube.dimension import Dimension, _Element, _Subtotal
from cr.cube.enums import DIMENSION_TYPE as DT
from cr.cube.matrix import (
    _AssembledVector,
    _BaseBaseMatrix,
    _BaseMatrixInsertedVector,
    _BaseTransformationVector,
    _BaseVector,
    _InsertedColumn,
    _InsertedRow,
    _OrderedVector,
    TransformedMatrix,
)

from ..unitutil import (
    ANY,
    call,
    class_mock,
    initializer_mock,
    instance_mock,
    method_mock,
    property_mock,
)


class DescribeTransformedMatrix(object):
    """Unit test suite for `cr.cube.matrix.TransformedMatrix` object."""

    def it_provides_a_constructor_classmethod(self, request, base_matrix_, cube_):
        _BaseBaseMatrix_ = class_mock(request, "cr.cube.matrix._BaseBaseMatrix")
        _BaseBaseMatrix_.factory.return_value = base_matrix_
        dimensions_ = tuple(instance_mock(request, Dimension) for _ in range(2))
        _init_ = initializer_mock(request, TransformedMatrix)

        matrix = TransformedMatrix.matrix(cube_, dimensions_, slice_idx=7)

        _BaseBaseMatrix_.factory.assert_called_once_with(cube_, dimensions_, 7)
        _init_.assert_called_once_with(ANY, base_matrix_)
        assert isinstance(matrix, TransformedMatrix)

    def it_assembles_inserted_columns_into_base_columns_to_help(
        self,
        _base_columns_prop_,
        _inserted_columns_prop_,
        _inserted_rows_prop_,
        _assembled_vectors_,
    ):
        _base_columns_prop_.return_value = ("base", "columns")
        _inserted_columns_prop_.return_value = ("inserted", "columns")
        _inserted_rows_prop_.return_value = ("inserted", "rows")
        _assembled_vectors_.return_value = ("assembled", "columns")
        matrix = TransformedMatrix(None)

        assembled_columns = matrix._assembled_columns

        _assembled_vectors_.assert_called_once_with(
            matrix, ("base", "columns"), ("inserted", "columns"), ("inserted", "rows")
        )
        assert assembled_columns == ("assembled", "columns")

    def it_assembles_inserted_rows_into_base_rows_to_help(
        self,
        _base_rows_prop_,
        _inserted_rows_prop_,
        _inserted_columns_prop_,
        _assembled_vectors_,
    ):
        _base_rows_prop_.return_value = ("base", "rows")
        _inserted_rows_prop_.return_value = ("inserted", "rows")
        _inserted_columns_prop_.return_value = ("inserted", "columns")
        _assembled_vectors_.return_value = ("assembled", "rows")
        matrix = TransformedMatrix(None)

        assembled_rows = matrix._assembled_rows

        _assembled_vectors_.assert_called_once_with(
            matrix, ("base", "rows"), ("inserted", "rows"), ("inserted", "columns")
        )
        assert assembled_rows == ("assembled", "rows")

    @pytest.mark.parametrize(
        ("InsertedVectorCls", "OpposingInsertedVectorCls"),
        ((_InsertedColumn, _InsertedRow), (_InsertedRow, _InsertedColumn)),
    )
    def it_assembles_base_vectors_with_inserted_vectors_to_help(
        self, request, InsertedVectorCls, OpposingInsertedVectorCls, _AssembledVector_
    ):
        # --- base vectors ---
        top_base_vector_ = instance_mock(request, _OrderedVector, name="top")
        top_base_vector_.ordering = (0, 0, top_base_vector_)
        bot_base_vector_ = instance_mock(request, _OrderedVector, name="bot")
        bot_base_vector_.ordering = (1, 1, bot_base_vector_)
        base_vectors_ = (top_base_vector_, bot_base_vector_)
        # --- inserted vectors ---
        top_inserted_vector_ = instance_mock(request, InsertedVectorCls, name="top")
        top_inserted_vector_.ordering = (0, -2, top_inserted_vector_)
        mid_inserted_vector_ = instance_mock(request, InsertedVectorCls, name="mid")
        mid_inserted_vector_.ordering = (1, -1, mid_inserted_vector_)
        bot_inserted_vector_ = instance_mock(request, InsertedVectorCls, name="bot")
        bot_inserted_vector_.ordering = (sys.maxsize, -3, bot_inserted_vector_)
        inserted_vectors_ = (
            bot_inserted_vector_,
            top_inserted_vector_,
            mid_inserted_vector_,
        )
        # --- opposing inserted vectors ---
        opposing_inserted_vectors_ = tuple(
            instance_mock(request, OpposingInsertedVectorCls) for _ in range(3)
        )
        # --- _AssembledVector behaviors ---
        assembled_vectors_ = tuple(
            instance_mock(request, _AssembledVector) for _ in range(8)
        )
        _AssembledVector_.side_effect = iter(assembled_vectors_)
        matrix = TransformedMatrix(None)

        assembled_vectors = matrix._assembled_vectors(
            base_vectors_, inserted_vectors_, opposing_inserted_vectors_
        )

        assert _AssembledVector_.call_args_list == [
            call(top_inserted_vector_, opposing_inserted_vectors_, 0),
            call(top_base_vector_, opposing_inserted_vectors_, 0),
            call(mid_inserted_vector_, opposing_inserted_vectors_, 0),
            call(bot_base_vector_, opposing_inserted_vectors_, 1),
            call(bot_inserted_vector_, opposing_inserted_vectors_, 0),
        ]
        assert assembled_vectors == assembled_vectors_[:5]

    def it_constructs_its_inserted_columns_to_help(
        self,
        request,
        _columns_dimension_prop_,
        dimension_,
        _base_rows_prop_,
        _base_columns_prop_,
        unordered_matrix_,
        _InsertedColumn_,
    ):
        subtotals_ = tuple(
            instance_mock(request, _Subtotal, name="subtot[%i]" % i) for i in range(3)
        )
        inserted_columns_ = tuple(
            instance_mock(request, _InsertedColumn) for _ in range(5)
        )
        dimension_.dimension_type = DT.CAT
        dimension_.subtotals = subtotals_
        _columns_dimension_prop_.return_value = dimension_
        _base_rows_prop_.return_value = ("base", "rows")
        _base_columns_prop_.return_value = ("base", "columns")
        unordered_matrix_.table_margin = 73
        _InsertedColumn_.side_effect = iter(inserted_columns_)
        matrix = TransformedMatrix(unordered_matrix_)

        inserted_columns = matrix._inserted_columns

        assert _InsertedColumn_.call_args_list == [
            call(subtotals_[0], -3, 73, ("base", "rows"), ("base", "columns")),
            call(subtotals_[1], -2, 73, ("base", "rows"), ("base", "columns")),
            call(subtotals_[2], -1, 73, ("base", "rows"), ("base", "columns")),
        ]
        assert inserted_columns == inserted_columns_[:3]

    @pytest.mark.parametrize("dimension_type", (DT.CA, DT.MR))
    def but_it_constructs_no_inserted_columns_for_an_array_dimension(
        self, dimension_type, _columns_dimension_prop_, dimension_
    ):
        dimension_.dimension_type = dimension_type
        _columns_dimension_prop_.return_value = dimension_
        matrix = TransformedMatrix(None)

        inserted_columns = matrix._inserted_columns

        assert inserted_columns == ()

    def it_constructs_its_inserted_rows_to_help(
        self,
        request,
        _rows_dimension_prop_,
        dimension_,
        _base_columns_prop_,
        _base_rows_prop_,
        unordered_matrix_,
        _InsertedRow_,
    ):
        subtotals_ = tuple(
            instance_mock(request, _Subtotal, name="subtot[%i]" % i) for i in range(3)
        )
        inserted_rows_ = tuple(instance_mock(request, _InsertedRow) for _ in range(5))
        dimension_.dimension_type = DT.CAT
        dimension_.subtotals = subtotals_
        _rows_dimension_prop_.return_value = dimension_
        _base_columns_prop_.return_value = ("base", "columns")
        _base_rows_prop_.return_value = ("base", "rows")
        unordered_matrix_.table_margin = 302
        _InsertedRow_.side_effect = iter(inserted_rows_)
        matrix = TransformedMatrix(unordered_matrix_)

        inserted_rows = matrix._inserted_rows

        assert _InsertedRow_.call_args_list == [
            call(subtotals_[0], -3, 302, ("base", "rows"), ("base", "columns")),
            call(subtotals_[1], -2, 302, ("base", "rows"), ("base", "columns")),
            call(subtotals_[2], -1, 302, ("base", "rows"), ("base", "columns")),
        ]
        assert inserted_rows == inserted_rows_[:3]

    @pytest.mark.parametrize("dimension_type", (DT.CA, DT.MR))
    def but_it_constructs_no_inserted_rows_for_an_array_dimension(
        self, dimension_type, _rows_dimension_prop_, dimension_
    ):
        dimension_.dimension_type = dimension_type
        _rows_dimension_prop_.return_value = dimension_
        matrix = TransformedMatrix(None)

        inserted_rows = matrix._inserted_rows

        assert inserted_rows == ()

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _AssembledVector_(self, request):
        return class_mock(request, "cr.cube.matrix._AssembledVector")

    @pytest.fixture
    def _assembled_vectors_(self, request):
        return method_mock(request, TransformedMatrix, "_assembled_vectors")

    @pytest.fixture
    def _base_columns_prop_(self, request):
        return property_mock(request, TransformedMatrix, "_base_columns")

    @pytest.fixture
    def base_matrix_(self, request):
        return instance_mock(request, _BaseBaseMatrix)

    @pytest.fixture
    def _base_rows_prop_(self, request):
        return property_mock(request, TransformedMatrix, "_base_rows")

    @pytest.fixture
    def _columns_dimension_prop_(self, request):
        return property_mock(request, TransformedMatrix, "_columns_dimension")

    @pytest.fixture
    def cube_(self, request):
        return instance_mock(request, Cube)

    @pytest.fixture
    def dimension_(self, request):
        return instance_mock(request, Dimension)

    @pytest.fixture
    def _InsertedColumn_(self, request):
        return class_mock(request, "cr.cube.matrix._InsertedColumn")

    @pytest.fixture
    def _inserted_columns_prop_(self, request):
        return property_mock(request, TransformedMatrix, "_inserted_columns")

    @pytest.fixture
    def _InsertedRow_(self, request):
        return class_mock(request, "cr.cube.matrix._InsertedRow")

    @pytest.fixture
    def _inserted_rows_prop_(self, request):
        return property_mock(request, TransformedMatrix, "_inserted_rows")

    @pytest.fixture
    def _rows_dimension_prop_(self, request):
        return property_mock(request, TransformedMatrix, "_rows_dimension")

    @pytest.fixture
    def unordered_matrix_(self, request):
        return instance_mock(request, _BaseBaseMatrix)


class Describe_BaseMatrixInsertedVector(object):
    """Unit test suite for `cr.cube.matrix._BaseMatrixInsertedVector` object."""

    def it_knows_the_indices_of_its_addends(
        self, request, subtotal_, _base_vectors_prop_
    ):
        subtotal_.addend_ids = (2, 3)
        _base_vectors_prop_.return_value = tuple(
            instance_mock(request, _BaseVector, element_id=i + 1) for i in range(6)
        )
        inserted_vector = _BaseMatrixInsertedVector(subtotal_, None, None, None, None)

        addend_idxs = inserted_vector.addend_idxs

        np.testing.assert_equal(addend_idxs, (1, 2))

    @pytest.mark.parametrize("anchor_value", ("top", "bottom", 42))
    def it_knows_its_anchor(self, subtotal_, anchor_value):
        subtotal_.anchor = anchor_value
        inserted_vector = _BaseMatrixInsertedVector(subtotal_, None, None, None, None)

        anchor = inserted_vector.anchor

        assert anchor == anchor_value

    def it_knows_it_is_inserted(self):
        assert (
            _BaseMatrixInsertedVector(None, None, None, None, None).is_inserted is True
        )

    def it_knows_its_ordering_value(self, request):
        property_mock(request, _BaseMatrixInsertedVector, "_anchor_n", return_value=8)
        inserted_vector = _BaseMatrixInsertedVector(None, -4, None, None, None)

        ordering = inserted_vector.ordering

        assert ordering == (8, -4, inserted_vector)

    @pytest.mark.parametrize(
        ("anchor", "expected_value"), (("top", 0), ("bottom", sys.maxsize), (3, 3))
    )
    def it_knows_its_anchor_location(
        self, request, anchor, expected_value, _base_vectors_prop_
    ):
        _base_vectors_prop_.return_value = tuple(
            instance_mock(request, _BaseVector, name="base[%d]" % i, element_id=i + 1)
            for i in range(5)
        )
        property_mock(request, _BaseMatrixInsertedVector, "anchor", return_value=anchor)
        inserted_vector = _BaseMatrixInsertedVector(None, None, None, None, None)

        anchor_n = inserted_vector._anchor_n

        assert anchor_n == expected_value

    @pytest.mark.parametrize("addend_idxs", ((), (1,), (2, 3)))
    def it_gathers_its_addend_vectors_to_help(
        self, request, addend_idxs, _base_vectors_prop_
    ):
        property_mock(
            request, _BaseMatrixInsertedVector, "addend_idxs", return_value=addend_idxs
        )
        base_vectors_ = tuple(
            instance_mock(request, _BaseVector, name="base[i]") for i in range(6)
        )
        _base_vectors_prop_.return_value = base_vectors_
        inserted_vector = _BaseMatrixInsertedVector(None, -4, None, None, None)

        addend_vectors = inserted_vector._addend_vectors

        assert addend_vectors == tuple(
            vector for i, vector in enumerate(base_vectors_) if i in addend_idxs
        )

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _base_vectors_prop_(self, request):
        return property_mock(request, _BaseMatrixInsertedVector, "_base_vectors")

    @pytest.fixture
    def subtotal_(self, request):
        return instance_mock(request, _Subtotal)


class Describe_BaseTransformationVector(object):
    """Unit test suite for `cr.cube.matrix._BaseTransformationVector` object."""

    def it_knows_its_element_id(self, base_vector_):
        base_vector_.element_id = 42
        transformation_vector = _BaseTransformationVector(base_vector_)

        element_id = transformation_vector.element_id

        assert element_id == 42

    # fixture components ---------------------------------------------

    @pytest.fixture
    def base_vector_(self, request):
        return instance_mock(request, _BaseVector)


class Describe_BaseVector(object):
    """Unit test suite for `cr.cube.matrix._BaseVector` object."""

    def it_knows_its_element_id(self, element_):
        element_.element_id = 42
        base_vector = _BaseVector(element_, None)

        element_id = base_vector.element_id

        assert element_id == 42

    # fixture components ---------------------------------------------

    @pytest.fixture
    def element_(self, request):
        return instance_mock(request, _Element)



# encoding: utf-8

"""Integration tests for pairwise comparisons."""

from unittest import TestCase

import numpy as np
import pytest

from cr.cube.cube import Cube
from cr.cube.cubepart import _Slice
from cr.cube.util import counts_with_subtotals

from ..fixtures import CR, SM
from ..util import load_python_expression


class Describe_Slice(object):
    """Integration-test suite for pairwise-t aspects of _Slice object."""

    @pytest.mark.parametrize(
        "fixture, pw_indices_dict, expectation",
        (
            (
                CR.PAIRWISE_HIROTSU_ILLNESS_X_OCCUPATION_WITH_HS,
                {},
                "cat-x-cat-hs-pw-idxs",
            ),
            (CR.PAIRWISE_HIROTSU_OCCUPATION_X_ILLNESS, {}, "cat-x-cat-pw-idxs"),
            (
                CR.PAIRWISE_HIROTSU_OCCUPATION_X_ILLNESS,
                {"only_larger": False},
                "cat-x-cat-pw-idxs-recip",
            ),
            (CR.MR_X_CAT, {}, "mr-x-cat-pw-idxs"),
            (CR.MR_X_CAT_HS, {}, "mr-x-cat-hs-pw-idxs"),
            (CR.EDU_FAV5_FAV5, {}, "cat-x-mr-aug-pw-idxs"),
            (CR.EDU_FAV5_FAV5, {"only_larger": False}, "cat-x-mr-aug-pw-idxs-recip"),
            (CR.CAT_HS_X_MR, {}, "cat-hs-x-mr-pw-idxs"),
            (CR.CAT_X_MR_2, {}, "cat-x-mr-pw-idxs"),
            (CR.MR_X_MR, {}, "mr-x-mr-pw-idxs"),
        ),
    )
    def it_provides_pairwise_indices(self, fixture, pw_indices_dict, expectation):
        slice_ = _Slice(
            Cube(fixture),
            slice_idx=0,
            transforms={"pairwise_indices": pw_indices_dict},
            population=None,
            mask_size=0,
        )

        actual = slice_.pairwise_indices

        expected = np.array(load_python_expression(expectation), dtype=tuple)
        assert (expected == actual).all(), "\n%s\n\n%s" % (expected, actual)

    @pytest.mark.parametrize(
        "fixture, pw_indices_dict, expectation",
        (
            (
                CR.PAIRWISE_HIROTSU_ILLNESS_X_OCCUPATION_WITH_HS,
                {"alpha": [0.175, 0.025], "only_larger": False},
                "cat-x-cat-hs-pw-idxs-alt",
            ),
            (CR.MR_X_CAT, {"alpha": [0.175, 0.01]}, "mr-x-cat-pw-idxs-alt"),
            (CR.EDU_FAV5_FAV5, {"alpha": [0.175, 0.01]}, "cat-x-mr-aug-pw-idxs-alt"),
        ),
    )
    def it_provides_pairwise_indices_alt(self, fixture, pw_indices_dict, expectation):
        """Provides indicies meeting secondary sig-test threshold, when specified."""
        slice_ = _Slice(
            Cube(fixture),
            slice_idx=0,
            transforms={"pairwise_indices": pw_indices_dict},
            population=None,
            mask_size=0,
        )

        actual = slice_.pairwise_indices_alt

        expected = np.array(load_python_expression(expectation), dtype=tuple)
        assert (expected == actual).all(), "\n%s\n\n%s" % (expected, actual)

    @pytest.mark.parametrize(
        "fixture, expectation",
        (
            (
                CR.PAIRWISE_HIROTSU_ILLNESS_X_OCCUPATION_WITH_HS,
                "cat-x-cat-hs-scale-mean-pw-idxs",
            ),
            (CR.PAIRWISE_HIROTSU_OCCUPATION_X_ILLNESS, "cat-x-cat-scale-mean-pw-idxs"),
            (CR.CAT_X_MR_2, "cat-x-mr-scale-mean-pw-idxs"),
            (CR.EDU_FAV5_FAV5, "cat-x-mr-aug-scale-mean-pw-idxs"),
        ),
    )
    def it_provides_scale_mean_pairwise_indices(self, fixture, expectation):
        """Provides column-indicies meeting sig-test threshold on column scale means."""
        slice_ = _Slice(
            Cube(fixture), slice_idx=0, transforms={}, population=None, mask_size=0
        )

        actual = slice_.scale_mean_pairwise_indices

        expected = load_python_expression(expectation)
        assert expected == actual, "\n%s\n\n%s" % (expected, actual)

    @pytest.mark.parametrize(
        "fixture, expectation",
        (
            (
                CR.PAIRWISE_HIROTSU_ILLNESS_X_OCCUPATION_WITH_HS,
                "cat-x-cat-hs-scale-mean-pw-idxs-alt",
            ),
            (
                CR.PAIRWISE_HIROTSU_OCCUPATION_X_ILLNESS,
                "cat-x-cat-scale-mean-pw-idxs-alt",
            ),
            (CR.CAT_X_MR_2, "cat-x-mr-scale-mean-pw-idxs-alt"),
            (CR.EDU_FAV5_FAV5, "cat-x-mr-aug-scale-mean-pw-idxs-alt"),
        ),
    )
    def it_provides_scale_mean_pairwise_indices_alt(self, fixture, expectation):
        """Provides col idxs meeting secondary sig-test threshold on scale mean."""
        slice_ = _Slice(
            Cube(fixture),
            slice_idx=0,
            transforms={"pairwise_indices": {"alpha": [0.01, 0.08]}},
            population=None,
            mask_size=0,
        )

        actual = slice_.scale_mean_pairwise_indices_alt

        expected = load_python_expression(expectation)
        assert expected == actual, "\n%s\n\n%s" % (expected, actual)


class TestStandardizedResiduals(TestCase):
    """Test cr.cube implementation of column family pairwise comparisons"""

    def test_ca_subvar_x_cat_hs_pairwise_t_tests(self):
        slice_ = Cube(CR.CA_SUBVAR_X_CAT_HS).partitions[0]
        actual = slice_.pairwise_significance_tests[1]

        np.testing.assert_array_almost_equal(
            actual.t_stats, load_python_expression("ca-subvar-x-cat-hs-pw-tstats")
        )
        np.testing.assert_array_almost_equal(
            actual.p_vals, load_python_expression("ca-subvar-x-cat-hs-pw-pvals")
        )

    def test_ca_subvar_hs_x_mr_augmented_pairwise_t_tests(self):
        slice_ = Cube(CR.CA_SUBVAR_HS_X_MR_AUGMENTED).partitions[0]
        actual = slice_.pairwise_significance_tests[1]
        overlap_margins = np.sum(slice_._cube.counts, axis=0)[:, 0, :, 0]
        addend_idxs = [s.addend_idxs for s in slice_._cube.dimensions[0].subtotals]
        counts_with_hs = counts_with_subtotals(
            addend_idxs, slice_.inserted_row_idxs, slice_._cube.counts
        )
        assert slice_.inserted_row_idxs == (0,)
        assert slice_.cube_is_mr_aug is True
        assert actual.t_stats.shape == (3, 4)
        assert slice_.counts.shape == (3, 4)
        assert counts_with_hs.shape == (3, 4, 2, 4, 2)
        np.testing.assert_array_almost_equal(
            overlap_margins,
            [[44, 3, 0, 0], [3, 34, 0, 0], [0, 0, 348, 0], [0, 0, 0, 0]],
        )

        slice_no_aug_ = Cube(CR.CA_SUBVAR_HS_X_MR).partitions[0]

        for i in range(slice_no_aug_.shape[0]):
            np.testing.assert_array_almost_equal(
                slice_no_aug_.counts[i], counts_with_hs[i][:, 0, :, 0].diagonal()
            )

    def test_cat_nps_numval_x_cat_scale_means_pariwise_t_tests(self):
        slice_ = Cube(SM.CAT_NPS_NUMVAL_X_CAT).partitions[0]

        actual = slice_.pairwise_significance_tests[0]
        assert actual.p_vals_scale_means[0] == 1
        assert actual.t_stats_scale_means[0] == 0
        np.testing.assert_almost_equal(
            actual.t_stats_scale_means, [0.0, 1.08966143, -1.27412884, -4.14629088]
        )
        np.testing.assert_almost_equal(
            actual.p_vals_scale_means,
            [1.0000000e00, 2.7638510e-01, 2.0302047e-01, 4.0226310e-05],
        )

        actual = slice_.pairwise_significance_tests[1]
        assert actual.p_vals_scale_means[1] == 1
        assert actual.t_stats_scale_means[1] == 0
        np.testing.assert_almost_equal(
            actual.p_vals_scale_means,
            [2.7638510e-01, 1.0000000e00, 1.4344585e-03, 1.7905266e-09],
        )
        np.testing.assert_almost_equal(
            actual.t_stats_scale_means, [-1.08966143, 0.0, -3.19741668, -6.10466696]
        )

    def test_cat_x_cat_pairwise_t_tests(self):
        slice_ = Cube(CR.PAIRWISE_HIROTSU_OCCUPATION_X_ILLNESS).partitions[0]
        actual = slice_.pairwise_significance_tests[2]

        np.testing.assert_almost_equal(
            actual.t_stats, load_python_expression("cat-x-cat-pw-tstats")
        )
        np.testing.assert_almost_equal(
            actual.p_vals, load_python_expression("cat-x-cat-pw-pvals")
        )

    def test_cat_x_cat_hs_pairwise_t_tests(self):
        slice_ = Cube(CR.PAIRWISE_HIROTSU_ILLNESS_X_OCCUPATION_WITH_HS).partitions[0]
        t_stats = slice_.pairwise_significance_tests[0].t_stats
        p_vals = slice_.pairwise_significance_tests[0].p_vals

        np.testing.assert_almost_equal(
            t_stats, load_python_expression("cat-x-cat-hs-pw-tstats")
        )
        np.testing.assert_almost_equal(
            p_vals, load_python_expression("cat-x-cat-hs-pw-pvals")
        )

    def test_cat_x_cat_hs_scale_means_pairwise_t_tests(self):
        slice_ = Cube(CR.PAIRWISE_HIROTSU_ILLNESS_X_OCCUPATION_WITH_HS).partitions[0]
        actual = slice_.pairwise_significance_tests[0]

        np.testing.assert_almost_equal(
            actual.t_stats_scale_means,
            load_python_expression("cat-x-cat-hs-scale-means-pw-tstats"),
        )
        np.testing.assert_almost_equal(
            actual.p_vals_scale_means,
            load_python_expression("cat-x-cat-hs-scale-means-pw-pvals"),
        )

    def test_cat_x_cat_pruning_and_hs_scale_means_pairwise_t_tests(self):
        transforms = {
            "columns_dimension": {"insertions": {}},
            "rows_dimension": {"insertions": {}},
        }
        slice_ = Cube(CR.CAT_X_CAT_PRUNING_HS, transforms=transforms).partitions[0]
        actual = slice_.pairwise_significance_tests[0]

        np.testing.assert_almost_equal(
            actual.t_stats_scale_means,
            [0.0, 1.64461503, 1.92387847, np.nan, 1.06912069],
        )
        np.testing.assert_almost_equal(
            actual.p_vals_scale_means, [1.0, 0.1046981, 0.059721, np.nan, 0.2918845]
        )

        # Just H&S
        slice_ = Cube(CR.CAT_X_CAT_PRUNING_HS).partitions[0]
        actual = slice_.pairwise_significance_tests[0]

        np.testing.assert_almost_equal(
            actual.t_stats_scale_means,
            [0.0, 0.9387958, 1.644615, 1.9238785, np.nan, 1.0691207],
        )

        np.testing.assert_almost_equal(
            actual.p_vals_scale_means,
            [1.0, 0.3500141, 0.1046981, 0.059721, np.nan, 0.2918845],
        )

        # Just pruning
        transforms = {
            "rows_dimension": {"prune": True},
            "columns_dimension": {"prune": True},
        }
        slice_ = Cube(CR.CAT_X_CAT_PRUNING_HS, transforms=transforms).partitions[0]
        actual = slice_.pairwise_significance_tests[0]

        np.testing.assert_almost_equal(
            actual.t_stats_scale_means,
            [0.0, 0.93879579, 1.64461503, 1.92387847, 1.06912069],
        )
        np.testing.assert_almost_equal(
            actual.p_vals_scale_means, [1.0, 0.3500141, 0.1046981, 0.059721, 0.2918845]
        )

        # Pruning and H&S
        transforms = {
            "rows_dimension": {"insertions": {}, "prune": True},
            "columns_dimension": {"insertions": {}, "prune": True},
        }
        slice_ = Cube(CR.CAT_X_CAT_PRUNING_HS, transforms=transforms).partitions[0]
        actual = slice_.pairwise_significance_tests[0]

        np.testing.assert_almost_equal(
            actual.t_stats_scale_means, [0.0, 1.64461503, 1.92387847, 1.06912069]
        )
        np.testing.assert_almost_equal(
            actual.p_vals_scale_means, [1.0, 0.1046981, 0.059721, 0.2918845]
        )

    def test_cat_x_cat_summary_pairwise_indices(self):
        # Only larger
        slice_ = Cube(CR.PAIRWISE_HIROTSU_OCCUPATION_X_ILLNESS).partitions[0]
        pairwise_indices = slice_.summary_pairwise_indices
        expected_indices = np.array([(2,), (0, 2), ()], dtype=tuple)
        np.testing.assert_array_equal(pairwise_indices, expected_indices)

        # Larger and smaller
        transforms = {"pairwise_indices": {"only_larger": False}}
        slice_ = Cube(
            CR.PAIRWISE_HIROTSU_OCCUPATION_X_ILLNESS, transforms=transforms
        ).partitions[0]
        pairwise_indices = slice_.summary_pairwise_indices
        expected_indices = np.array([(1, 2), (0, 2), (0, 1)], dtype="i,i")
        np.testing.assert_array_equal(pairwise_indices, expected_indices)

    def test_cat_x_cat_wgtd_pairwise_t_tests(self):
        """The weights on this cube demonstrate much higher variance (less
        extreme t values, and higher associated p-values) than if weighted_n
        were used in the variance estimate of the test statistic.
        """
        slice_ = Cube(CR.CAT_X_CAT_WEIGHTED_TTESTS).partitions[0]
        actual = slice_.pairwise_significance_tests[0]
        pairwise_indices = slice_.pairwise_indices

        np.testing.assert_almost_equal(
            actual.t_stats, load_python_expression("cat-x-cat-wgtd-pw-tstats")
        )
        np.testing.assert_almost_equal(
            actual.p_vals, load_python_expression("cat-x-cat-wgtd-pw-pvals")
        )
        np.testing.assert_array_equal(
            pairwise_indices,
            np.array(load_python_expression("cat-x-cat-wgtd-pw-indices"), dtype=tuple),
        )

    def test_cat_x_cat_wgtd_scale_means_pariwise_t_tests(self):
        slice_ = Cube(CR.CAT_X_CAT_WEIGHTED_TTESTS).partitions[0]
        actual = slice_.pairwise_significance_tests[0]

        np.testing.assert_almost_equal(
            actual.t_stats_scale_means, [0.0, -4.38871748, -3.99008596, -5.15679647]
        )
        np.testing.assert_almost_equal(
            actual.p_vals_scale_means,
            [1.0000000e00, 1.3839564e-05, 7.4552516e-05, 4.0145665e-07],
        )

    def test_cat_hs_x_mr_wgtd_augmented_pairwise_t_tests(self):
        slice_ = Cube(CR.CAT_HS_X_MR_AUGMENTED_WGTD).partitions[0]
        actual = slice_.pairwise_significance_tests[1]

        assert slice_.cube_is_mr_aug is True
        np.testing.assert_array_almost_equal(
            actual.t_stats, load_python_expression("cat-hs-x-mr-wgtd-aug-pw-tstats")
        )
        np.testing.assert_array_almost_equal(
            actual.p_vals, load_python_expression("cat-hs-x-mr-wgtd-aug-pw-pvals")
        )

    def test_cat_x_mr_weighted_augmented_pairwise_t_tests(self):
        slice_ = Cube(CR.CAT_X_MR_WEIGHTED_AUGMENTED).partitions[0]
        actual = slice_.pairwise_significance_tests[1]
        overlap_margins = np.sum(slice_._cube.counts, axis=0)[:, 0, :, 0]
        shadow_proportions_tab1 = slice_._cube.counts[0][:, 0, :, 0] / overlap_margins
        shadow_proportions_tab2 = slice_._cube.counts[1][:, 0, :, 0] / overlap_margins

        assert slice_.cube_is_mr_aug is True
        # --- each diagonal of the shadow proportions tab is equal to the correspondent
        # --- row for the actual proportions
        np.testing.assert_array_almost_equal(
            shadow_proportions_tab1.diagonal(), slice_.column_proportions[0, :]
        )
        np.testing.assert_array_almost_equal(
            shadow_proportions_tab2.diagonal(), slice_.column_proportions[1, :]
        )
        np.testing.assert_array_almost_equal(
            actual.t_stats,
            [[-10.2964264, 0.0, -20.09577285], [5.96350953, 0.0, 24.14335882]],
        )
        np.testing.assert_array_almost_equal(
            actual.p_vals,
            [[0.0, 1.00000000e00, 0.0], [2.55612775e-09, 1.00000000e00, 0.0]],
        )

    def test_mr_x_mr_pairwise_t_tests(self):
        slice_ = Cube(CR.MR_X_MR_2).partitions[0]
        actual = slice_.pairwise_significance_tests[1]

        assert slice_.cube_is_mr_aug is False
        np.testing.assert_array_almost_equal(
            actual.t_stats, load_python_expression("mr-x-mr-pw-tstats")
        )
        np.testing.assert_array_almost_equal(
            actual.p_vals, load_python_expression("mr-x-mr-pw-pvals")
        )

    def test_mr_x_mr_weighted_augmented_pairwise_t_tests(self):
        """This test proofs the hypotesis testing for MR1_X_MR2 considering the
        overlaps. To calculate the overlap of this kind of cube we need to calculate
        the shadow proportion behind the MR1_X_MR2 table:
                            +-----------------------------------+
                            |               Trust               |
        +-------------------+-----------------------------------+
        | Pol.Know          |                                   |
        +-------------------+----------- +---------+------------+
        |                   |    NYTimes |  WaPo   |  FoxNews   |
        +===================+============+=========+============+
        | Senate            |    71.39   |  74.80  |    55.48   |
        | Deficit           |    67.75   |  72.58  |    44.91   |
        | Unemp.            |    62.70   |  66.04  |    68.24   |
        | Tariffs           |    80.47   |  83.61  |    64.52   |
        | Elec.vot          |    43.52   |  44.56  |    33.81   |
        +-------------------+------------+---------+------------+

         ==================SENATE - SELECTED (COUNTS)==============
                            +-----------------------------------+
                            |               Trust               |
        +-------------------+-----------------------------------+
        | Trust             |                                   |
        +-------------------+----------- +---------+------------+
        |                   |    NYTimes |  WaPo   |  FoxNews   |
        +===================+============+=========+============+
        | NYTimes           |    2990    |  2309   |    461     |
        | WaPo              |    2309    |  2714   |    428     |
        | FoxNews           |    461     |  428    |    2848    |
        +-------------------+------------+---------+------------+

        ==================SENATE - OTHER (COUNTS)================
                            +-----------------------------------+
                            |               Trust               |
        +-------------------+-----------------------------------+
        | Trust             |                                   |
        +-------------------+----------- +---------+------------+
        |                   |    NYTimes |  WaPo   |  FoxNews   |
        +===================+============+=========+============+
        | NYTimes           |   1998     |  737    |    586     |
        | WaPo              |    737     |  914    |    431     |
        | FoxNews           |    586     |  431    |    2285    |
        +-------------------+------------+---------+------------+

        So to get the 71.39 we have to divide counts / margin. For example
        the first shadow proportion will be 2990 / (2990+1198) where 2990 is the count
        of the NYTimesXNYTimes in the Senate Selected tab and the denominator is the
        sum of the counts of the Selected and Other Senate tab in the position (0,0).

        This procedure, is repeated for each couple of (selected and other) tabs for
        each of the 5 Pol.Know variables.

        Given the shadow proportions and the overlap margins for all the tabs we use
        these figures to calculate the t-statistic considering the overlaps.

        In this test we are comparing the first column `NYTimes` with the other 2
        using the correct t_stats arrays."""
        slice_ = Cube(CR.MR_X_MR_WEIGHTED_AUGMENTED).partitions[0]
        actual = slice_.pairwise_significance_tests[1]
        overlap_margins = np.sum(slice_._cube.counts[0], axis=0)[:, 0, :, 0]
        # MRxMRxMR has (5, 2, 3, 2, 3, 2) shape, so 5 tabs of shadow proportions
        shadow_proportions_tab = [
            (slice_._cube.counts[i][0][:, 0, :, 0] / overlap_margins)
            for i in range(slice_._cube.counts.shape[0])
        ]

        assert slice_.cube_is_mr_aug is True
        np.testing.assert_array_almost_equal(
            actual.t_stats, load_python_expression("mr-x-mr-wgtd-aug-pw-tstats")
        )
        np.testing.assert_array_almost_equal(
            actual.p_vals, load_python_expression("mr-x-mr-wgtd-aug-pw-pvals")
        )
        # each diagonal of the shadow proportions tab is equal to the correspondent row
        # for the actual proportions
        np.testing.assert_array_almost_equal(
            shadow_proportions_tab[0].diagonal(), slice_.column_proportions[0, :]
        )
        np.testing.assert_array_almost_equal(
            shadow_proportions_tab[1].diagonal(), slice_.column_proportions[1, :]
        )
        np.testing.assert_array_almost_equal(
            shadow_proportions_tab[2].diagonal(), slice_.column_proportions[2, :]
        )
        np.testing.assert_array_almost_equal(
            shadow_proportions_tab[3].diagonal(), slice_.column_proportions[3, :]
        )
        np.testing.assert_array_almost_equal(
            shadow_proportions_tab[4].diagonal(), slice_.column_proportions[4, :]
        )
        assert slice_._cube.counts.shape == (5, 2, 3, 2, 3, 2)
        assert actual.t_stats.shape == (5, 3)

    def test_mr_subvar_x_mr_augmented_pairwise_t_tests(self):
        """This test proofs the hypotesis testing for MR_SUBVAR_X_MR considering the
        overlaps. To calculate the overlap of this kind of cube we need to calculate
        the shadow proportion behind the SUBVAR_X_MR table:
                            +-----------------------------------+
                            |               Trust               |
        +-------------------+-----------------------------------+
        | Increasing Taxes  |                                   |
        +-------------------+----------- +---------+------------+
        |                   |    NYTimes |  WaPo   |  FoxNews   |
        +===================+============+=========+============+
        | Selected          |    82.71   |  85.96  |    63.21   |
        | Other             |    17.29   |  14.4   |    36.79   |
        +-------------------+------------+---------+------------+

         ==================TAB 1 - SELECTED (COUNTS)=============
                            +-----------------------------------+
                            |               Trust               |
        +-------------------+-----------------------------------+
        | Trust             |                                   |
        +-------------------+----------- +---------+------------+
        |                   |    NYTimes |  WaPo   |  FoxNews   |
        +===================+============+=========+============+
        | NYTimes           |    3464    |  2647   |    590     |
        | WaPo              |    2647    |  3119   |    529     |
        | FoxNews           |    590     |  529    |    3245    |
        +-------------------+------------+---------+------------+

        ==================TAB 2 - OTHER (COUNTS)=================
                            +-----------------------------------+
                            |               Trust               |
        +-------------------+-----------------------------------+
        | Trust             |                                   |
        +-------------------+----------- +---------+------------+
        |                   |    NYTimes |  WaPo   |  FoxNews   |
        +===================+============+=========+============+
        | NYTimes           |    724     |  400    |    458     |
        | WaPo              |    400     |  509    |    331     |
        | FoxNews           |    458     |  331    |    1889    |
        +-------------------+------------+---------+------------+

        So to get the 82.71 we have to divide counts / margin. For example
        the first shadow proportion will be 3464 / (3464+724) where 3464 is the count
        of the NYTimesXNYTimes in the first tab and the denominator is the sum of the
        counts of the first tab and second tab in the position (0,0).

        Given the shadow proportions and the overlap margins for all the tabs we use
        these figures to calculate the t-statistic considering the overlaps.

        In this test we are comparing the first column `NYTimes` with the other 2
        using the correct t_stats arrays.
        """
        slice_ = Cube(CR.CAT_SUBVAR_X_MR_AUGMENTED).partitions[0]
        actual = slice_.pairwise_significance_tests[0]
        overlap_margins = np.sum(slice_._cube.counts, axis=0)[:, 0, :, 0]
        shadow_proportions_tab1 = slice_._cube.counts[0][:, 0, :, 0] / overlap_margins
        shadow_proportions_tab2 = slice_._cube.counts[1][:, 0, :, 0] / overlap_margins

        assert slice_.cube_is_mr_aug is True
        np.testing.assert_array_almost_equal(
            actual.t_stats, load_python_expression("mr-subvar-x-mr-aug-pw-tstats")
        )
        np.testing.assert_array_almost_equal(
            actual.p_vals, load_python_expression("mr-subvar-x-mr-aug-pw-pvals")
        )
        np.testing.assert_array_almost_equal(
            shadow_proportions_tab1,
            load_python_expression("mr-subvar-x-mr-aug-pw-shadow-proportions-tab1"),
        )
        np.testing.assert_array_almost_equal(
            shadow_proportions_tab2,
            load_python_expression("mr-subvar-x-mr-aug-pw-shadow-proportions-tab2"),
        )
        np.testing.assert_array_almost_equal(
            shadow_proportions_tab1.diagonal(), slice_.column_proportions[0, :]
        )
        np.testing.assert_array_almost_equal(
            shadow_proportions_tab2.diagonal(), slice_.column_proportions[1, :]
        )



# encoding: utf-8

"""T-score based P-values of pairwise comparison or columns of a contingency table."""

from __future__ import division

import numpy as np
from scipy.stats import t

from cr.cube.util import lazyproperty

try:
    xrange
except NameError:  # pragma: no cover
    # pylint: disable=invalid-name
    xrange = range


class PairwiseSignificance:
    """Implementation of p-vals and t-tests for each column proportions comparison."""

    def __init__(
        self, slice_, axis=0, weighted=True, alpha=0.05, only_larger=True, hs_dims=None
    ):
        self._slice = slice_
        self._axis = axis
        self._weighted = weighted
        self._alpha = alpha
        self._only_larger = only_larger
        self._hs_dims = hs_dims

    @lazyproperty
    def values(self):
        """list of _ColumnPairwiseSignificance tests.

        Result has as many elements as there are coliumns in the slice. Each
        significance test contains `p_vals` and `t_stats` significance tests.
        """
        # TODO: Figure out how to intersperse pairwise objects for columns
        # that represent H&S
        return [
            _ColumnPairwiseSignificance(
                self._slice,
                col_idx,
                self._axis,
                self._weighted,
                self._alpha,
                self._only_larger,
                self._hs_dims,
            )
            for col_idx in range(self._slice.get_shape(hs_dims=self._hs_dims)[1])
        ]

    @lazyproperty
    def pairwise_indices(self):
        """ndarray containing tuples of pairwise indices."""
        return np.array([sig.pairwise_indices for sig in self.values], dtype=object).T

    @lazyproperty
    def summary_pairwise_indices(self):
        """ndarray containing tuples of pairwise indices for the column summary."""
        summary_pairwise_indices = np.empty(
            self.values[0].t_stats.shape[1], dtype=object
        )
        summary_pairwise_indices[:] = [
            sig.summary_pairwise_indices for sig in self.values
        ]
        return summary_pairwise_indices


# pylint: disable=too-few-public-methods
class _ColumnPairwiseSignificance:
    """Value object providing matrix of T-score based pairwise-comparison P-values"""

    def __init__(
        self,
        slice_,
        col_idx,
        axis=0,
        weighted=True,
        alpha=0.05,
        only_larger=True,
        hs_dims=None,
    ):
        self._slice = slice_
        self._col_idx = col_idx
        self._axis = axis
        self._weighted = weighted
        self._alpha = alpha
        self._only_larger = only_larger
        self._hs_dims = hs_dims

    @lazyproperty
    def _unweighted_col_margin(self):
        return self._slice.margin(
            axis=0, weighted=False, include_transforms_for_dims=self._hs_dims
        )

    @lazyproperty
    def t_stats(self):
        props = self._slice.proportions(
            axis=0, include_transforms_for_dims=self._hs_dims
        )
        diff = props - props[:, [self._col_idx]]
        var_props = props * (1.0 - props) / self._unweighted_col_margin
        se_diff = np.sqrt(var_props + var_props[:, [self._col_idx]])
        return diff / se_diff

    @lazyproperty
    def p_vals(self):
        return 2 * (1 - t.cdf(abs(self.t_stats), df=self._df))

    @lazyproperty
    def pairwise_indices(self):
        significance = self.p_vals < self._alpha
        if self._only_larger:
            significance = np.logical_and(self.t_stats < 0, significance)
        return [tuple(np.where(sig_row)[0]) for sig_row in significance]

    @lazyproperty
    def summary_pairwise_indices(self):
        significance = self.summary_p_vals < self._alpha
        if self._only_larger:
            significance = np.logical_and(self.summary_t_stats < 0, significance)
        return tuple(np.where(significance)[0])

    @lazyproperty
    def summary_t_stats(self):
        total_margin = self._slice.margin(weighted=self._weighted)
        col_margin_props = self._unweighted_col_margin / total_margin
        diff = col_margin_props - col_margin_props[self._col_idx]
        var_props = col_margin_props * (1.0 - col_margin_props) / total_margin
        se_diff = np.sqrt(var_props + var_props[self._col_idx])
        return diff / se_diff

    @lazyproperty
    def summary_p_vals(self):
        return 2 * (1 - t.cdf(abs(self.summary_t_stats), df=self._df))

    @lazyproperty
    def _df(self):
        selected_unweighted_n = (
            self._unweighted_n[self._col_idx]
            if self._unweighted_n.ndim < 2
            else self._unweighted_n[:, self._col_idx][:, None]
        )
        return self._unweighted_n + selected_unweighted_n - 2

    @lazyproperty
    def _unweighted_n(self):
        return self._slice.margin(
            axis=0, weighted=False, include_transforms_for_dims=self._hs_dims
        )



import numpy as np

from cr.cube.cube import Cube

from ..fixtures import CR, SM  # ---mnemonic: SM = 'scale means'---


def test_ca_cat_x_items():
    slice_ = Cube(SM.CA_CAT_X_ITEMS).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row, [1.60128358, 2.27572817, 2.1804707, 2.19456437]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row, [0.0394328, 0.0579909, 0.0537937, 0.0544237]
    )
    np.testing.assert_almost_equal(slice_.scale_median_row, [1, 5, 5, 5])
    assert slice_.scale_median_row_margin == 4
    assert slice_.scale_median_column_margin is None
    assert slice_.scale_std_dev_column is None
    assert slice_.scale_std_err_column is None
    assert slice_.scale_median_column is None


def test_ca_items_x_cat():
    slice_ = Cube(SM.CA_ITEMS_X_CAT).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_column, [1.60128358, 2.27572817, 2.1804707, 2.19456437]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_column, [0.0394328, 0.0579909, 0.0537937, 0.0544237]
    )
    np.testing.assert_almost_equal(slice_.scale_median_column, [1, 5, 5, 5])
    assert slice_.scale_median_column_margin == 4
    assert slice_.scale_median_row_margin is None
    assert slice_.scale_std_dev_row is None
    assert slice_.scale_std_err_row is None
    assert slice_.scale_median_row is None


def test_ca_x_mr():
    slice_ = Cube(SM.CA_X_MR).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row, [1.43286986, 1.75336411, 1.59143099, np.nan]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row, [0.1206694, 0.1481863, 0.0428554, np.nan]
    )
    np.testing.assert_almost_equal(slice_.scale_median_row, [1, 1, 1, np.nan])
    assert slice_.scale_median_row_margin == 1
    assert slice_.scale_median_column_margin is None
    assert slice_.scale_std_dev_column is None
    assert slice_.scale_std_err_column is None
    assert slice_.scale_median_column is None

    slice_ = Cube(SM.CA_X_MR).partitions[1]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row, [2.23837666, 2.32059287, 2.2735136, np.nan]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row, [0.1994104, 0.210963, 0.0629833, np.nan]
    )
    np.testing.assert_almost_equal(slice_.scale_median_row, [5.0, 5.0, 4.0, np.nan])
    assert slice_.scale_median_row_margin == 5
    assert slice_.scale_median_column_margin is None
    assert slice_.scale_std_dev_column is None
    assert slice_.scale_std_err_column is None
    assert slice_.scale_median_column is None

    slice_ = Cube(SM.CA_X_MR).partitions[2]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row, [2.24756685, 2.18544045, 2.17207083, np.nan]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row, [0.1892793, 0.186037, 0.0585764, np.nan]
    )
    np.testing.assert_almost_equal(slice_.scale_median_row, [5.0, 5.0, 5.0, np.nan])
    assert slice_.scale_median_row_margin == 5
    assert slice_.scale_median_column_margin is None
    assert slice_.scale_std_dev_column is None
    assert slice_.scale_std_err_column is None
    assert slice_.scale_median_column is None

    slice_ = Cube(SM.CA_X_MR).partitions[3]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row, [2.1892387, 2.2020816, 2.1932015, np.nan]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row, [0.1877257, 0.189525, 0.0593408, np.nan]
    )
    np.testing.assert_almost_equal(slice_.scale_median_row, [5.0, 5.0, 5.0, np.nan])
    assert slice_.scale_median_row_margin == 5
    assert slice_.scale_median_column_margin is None
    assert slice_.scale_std_dev_column is None
    assert slice_.scale_std_err_column is None
    assert slice_.scale_median_column is None


def test_cat_x_ca_cat_x_items():
    slice_ = Cube(SM.CAT_X_CA_CAT_X_ITEMS).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row, [1.6092707, 2.241556, 2.22469, 2.2660994]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row, [0.216994, 0.3202223, 0.302742, 0.3055608]
    )
    np.testing.assert_almost_equal(slice_.scale_median_row, [1, 1, 2.5, 1])
    assert slice_.scale_median_row_margin == 1
    assert slice_.scale_median_column_margin is None
    assert slice_.scale_std_dev_column is None
    assert slice_.scale_std_err_column is None
    assert slice_.scale_median_column is None

    slice_ = Cube(SM.CAT_X_CA_CAT_X_ITEMS).partitions[1]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row, [1.661212, 2.2292455, 2.1188676, 2.095007]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row, [0.1491812, 0.2097098, 0.1895173, 0.1889004]
    )
    np.testing.assert_almost_equal(slice_.scale_median_row, [1, 5, 5, 5])
    assert slice_.scale_median_row_margin == 4
    assert slice_.scale_median_column_margin is None
    assert slice_.scale_std_dev_column is None
    assert slice_.scale_std_err_column is None
    assert slice_.scale_median_column is None

    slice_ = Cube(SM.CAT_X_CA_CAT_X_ITEMS).partitions[2]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row, [1.5301377, 2.172702, 2.1050968, 2.1110668]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row, [0.0618017, 0.0902944, 0.0852329, 0.0857562]
    )
    np.testing.assert_almost_equal(slice_.scale_median_row, [1, 5, 5, 5])
    assert slice_.scale_median_row_margin == 4
    assert slice_.scale_median_column_margin is None
    assert slice_.scale_std_dev_column is None
    assert slice_.scale_std_err_column is None
    assert slice_.scale_median_column is None

    slice_ = Cube(SM.CAT_X_CA_CAT_X_ITEMS).partitions[3]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row, [1.5764139, 2.3345044, 2.1969246, 2.2629614]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row, [0.0896791, 0.1385273, 0.1251814, 0.1308704]
    )
    np.testing.assert_almost_equal(slice_.scale_median_row, [1, 5, 5, 5])
    assert slice_.scale_median_row_margin == 4
    assert slice_.scale_median_column_margin is None
    assert slice_.scale_std_dev_column is None
    assert slice_.scale_std_err_column is None
    assert slice_.scale_median_column is None

    slice_ = Cube(SM.CAT_X_CA_CAT_X_ITEMS).partitions[4]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row, [1.5302137, 2.4273602, 2.3953561, 2.3938526]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row, [0.0767992, 0.124521, 0.1202195, 0.1204478]
    )
    np.testing.assert_almost_equal(slice_.scale_median_row, [1, 4, 4, 4])
    assert slice_.scale_median_row_margin == 1
    assert slice_.scale_median_column_margin is None
    assert slice_.scale_std_dev_column is None
    assert slice_.scale_std_err_column is None
    assert slice_.scale_median_column is None

    slice_ = Cube(SM.CAT_X_CA_CAT_X_ITEMS).partitions[5]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row, [1.6549498, 1.656014, 1.3338366, 1.3250429]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row, [0.136498, 0.1435946, 0.1107691, 0.1104202]
    )
    assert slice_.scale_median_row_margin == 4
    assert slice_.scale_median_column_margin is None
    assert slice_.scale_std_dev_column is None
    assert slice_.scale_std_err_column is None
    assert slice_.scale_median_column is None


def test_cat_x_cat():
    slice_ = Cube(SM.CAT_X_CAT).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_column,
        [1.83663364, 1.85424071, 1.85212099, 1.78435798, 1.78496323, 1.8552804],
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row, [2.35648861, 2.32601576, 2.3064134, 2.14673544]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_column,
        [0.2476516, 0.1651889, 0.0748065, 0.1013447, 0.0892482, 0.1525031],
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row, [0.1135081, 0.1046524, 0.1345128, 0.1031655]
    )
    np.testing.assert_almost_equal(slice_.scale_median_row, [2, 2, 2, 4])
    np.testing.assert_almost_equal(slice_.scale_median_column, [0, 2, 2, 2, 4, 2])
    # Test Scale Variance
    np.testing.assert_almost_equal(
        slice_.var_scale_means_row, [5.55303858, 5.41034929, 5.31954278, 4.60847303]
    )
    np.testing.assert_almost_equal(
        slice_.var_scale_means_column,
        [3.37322314, 3.43820862, 3.43035216, 3.1839334, 3.18609375, 3.44206538],
    )
    assert slice_.scale_median_row_margin == 4
    assert slice_.scale_median_column_margin == 2


def test_cat_x_mr():
    slice_ = Cube(SM.CAT_X_MR).partitions[0]

    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row, [1.84458654, 1.92650335, 1.83760589, np.nan]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row, [0.1547943, 0.173707, 0.0493952, np.nan]
    )
    np.testing.assert_almost_equal(slice_.scale_median_row, [2.0, 2.0, 2.0, np.nan])
    assert slice_.scale_median_row_margin == 2
    assert slice_.scale_median_column_margin is None
    assert slice_.scale_std_dev_column is None
    assert slice_.scale_std_err_column is None
    assert slice_.scale_median_column is None


def test_mr_x_cat():
    slice_ = Cube(SM.MR_X_CAT).partitions[0]

    np.testing.assert_almost_equal(
        slice_.scale_std_dev_column, [1.84458654, 1.92650335, 1.83760589, np.nan]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_column, [0.1547943, 0.173707, 0.0493952, np.nan]
    )
    np.testing.assert_almost_equal(slice_.scale_median_column, [2.0, 2.0, 2.0, np.nan])
    assert slice_.scale_median_column_margin == 2
    assert slice_.scale_median_row_margin is None
    assert slice_.scale_std_dev_row is None
    assert slice_.scale_std_err_row is None
    assert slice_.scale_median_row is None


def test_univariate_cat():
    strand_ = Cube(SM.UNIVARIATE_CAT).partitions[0]
    np.testing.assert_almost_equal(strand_.scale_std_dev, [2.3307207194301434])
    np.testing.assert_almost_equal(strand_.scale_std_err, [0.05755304590263942])
    np.testing.assert_almost_equal(strand_.scale_median, [4.0])


def test_cat_x_cat_without_scales():
    strand_ = Cube(SM.UNIVARIATE_CAT_WITH_ALL_NULL_NUMERIC_VALUE).partitions[0]
    assert strand_.scale_median is None
    assert strand_.scale_std_dev is None
    assert strand_.scale_std_err is None


def test_univariate_with_hs():
    # Test without H&S
    transforms = {
        "columns_dimension": {"insertions": {}},
        "rows_dimension": {"insertions": {}},
    }
    strand_ = Cube(CR.ECON_BLAME_WITH_HS, transforms).partitions[0]
    np.testing.assert_almost_equal(strand_.scale_std_dev, [1.066016])
    np.testing.assert_almost_equal(strand_.scale_std_err, [0.0337611])
    np.testing.assert_almost_equal(strand_.scale_median, [2.0])

    # Test with H&S
    strand_ = Cube(CR.ECON_BLAME_WITH_HS).partitions[0]
    np.testing.assert_almost_equal(strand_.scale_std_dev, [1.066016])
    np.testing.assert_almost_equal(strand_.scale_std_err, [0.0337611])
    np.testing.assert_almost_equal(strand_.scale_median, [2.0])


def test_univariate_cat_with_hiding():
    strand_ = Cube(SM.BOLSHEVIK_HAIR).partitions[0]
    np.testing.assert_almost_equal(strand_.scale_std_dev, [1.6012836])
    np.testing.assert_almost_equal(strand_.scale_std_err, [0.0394328])
    np.testing.assert_almost_equal(strand_.scale_median, [1.0])

    # Appling hiding transforms
    transforms = {
        "rows_dimension": {"elements": {"5": {"hide": True}, "4": {"hide": True}}}
    }
    strand_with_hiding_ = Cube(SM.BOLSHEVIK_HAIR, transforms=transforms).partitions[0]

    np.testing.assert_almost_equal(
        strand_.scale_std_dev, strand_with_hiding_.scale_std_dev
    )
    np.testing.assert_almost_equal(
        strand_.scale_std_err, strand_with_hiding_.scale_std_err
    )
    np.testing.assert_almost_equal(strand_.scale_median, [1.0])


def test_cat_x_cat_with_hs():
    # Test without H&S
    transforms = {
        "columns_dimension": {"insertions": {}},
        "rows_dimension": {"insertions": {}},
    }
    slice_ = Cube(CR.ECON_BLAME_X_IDEOLOGY_ROW_HS, transforms=transforms).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_column,
        [0.943031, 0.9677583, 0.9817768, 1.8856181, 1.5987533],
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row,
        [0.7195463, 0.7196963, 0.9977753, 1.0608932, 1.0948414, 1.5740076],
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_column,
        [0.0558603, 0.0486317, 0.063111, 0.7698004, 0.1938773],
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row,
        [0.0847993, 0.0533474, 0.0515249, 0.0718528, 0.104389, 0.2488725],
    )
    np.testing.assert_almost_equal(slice_.scale_median_row, [2, 2, 2, 1, 1, 5])
    np.testing.assert_almost_equal(slice_.scale_median_column, [4, 3, 3, 3.5, 4])
    assert slice_.scale_median_row_margin == 2
    assert slice_.scale_median_column_margin == 3

    # Test with H&S
    slice_ = Cube(CR.ECON_BLAME_X_IDEOLOGY_ROW_HS).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_column,
        [0.943031, 0.9677583, 1.1680149, 0.9817768, 1.8856181, 1.5987533],
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row,
        [0.7195463, 0.7196963, 0.9977753, 1.0608932, 1.0948414, 1.5740076],
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_column,
        [0.0558603, 0.0486317, 0.0447584, 0.063111, 0.7698004, 0.1938773],
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row,
        [0.0847993, 0.0533474, 0.0515249, 0.0718528, 0.104389, 0.2488725],
    )
    np.testing.assert_almost_equal(slice_.scale_median_row, [2, 2, 2, 1, 1, 5])
    np.testing.assert_almost_equal(slice_.scale_median_column, [4, 3, 3, 3, 3.5, 4])
    assert slice_.scale_median_row_margin == 2
    assert slice_.scale_median_column_margin == 3


def test_cat_x_cat_with_hs_on_both_dims():
    # Test without H&S
    transforms = {
        "columns_dimension": {"insertions": {}},
        "rows_dimension": {"insertions": {}},
    }
    slice_ = Cube(
        CR.ECON_BLAME_X_IDEOLOGY_ROW_AND_COL_HS, transforms=transforms
    ).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_column,
        [0.943031, 0.9677583, 0.9817768, 1.8856181, 1.5987533],
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row,
        [0.7195463, 0.7196963, 0.9977753, 1.0608932, 1.0948414, 1.5740076],
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_column,
        [0.0558603, 0.0486317, 0.063111, 0.7698004, 0.1938773],
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row,
        [0.0847993, 0.0533474, 0.0515249, 0.0718528, 0.104389, 0.2488725],
    )
    np.testing.assert_almost_equal(slice_.scale_median_row, [2, 2, 2, 1, 1, 5])
    np.testing.assert_almost_equal(slice_.scale_median_column, [4, 3, 3, 3.5, 4])
    assert slice_.scale_median_row_margin == 2
    assert slice_.scale_median_column_margin == 3

    # Test with H&S
    slice_ = Cube(CR.ECON_BLAME_X_IDEOLOGY_ROW_AND_COL_HS).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_column,
        [0.943031, 0.9677583, 1.1680149, 0.9817768, 1.8856181, 1.5987533],
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row,
        [0.7195463, 0.7196963, 0.9977753, 0.9169069, 1.0608932, 1.0948414, 1.5740076],
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_column,
        [0.0558603, 0.0486317, 0.0447584, 0.063111, 0.7698004, 0.1938773],
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row,
        [0.0847993, 0.0533474, 0.0515249, 0.0388506, 0.0718528, 0.104389, 0.2488725],
    )
    np.testing.assert_almost_equal(slice_.scale_median_row, [2, 2, 2, 2, 1, 1, 5])
    np.testing.assert_almost_equal(slice_.scale_median_column, [4, 3, 3, 3, 3.5, 4])
    assert slice_.scale_median_row_margin == 2
    assert slice_.scale_median_column_margin == 3


def test_ca_x_mr_with_hs_and_pruning():
    transforms = {
        "columns_dimension": {"insertions": {}},
        "rows_dimension": {"insertions": {}},
    }
    slice_ = Cube(CR.CA_X_MR_HS, transforms=transforms).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row, [1.4689114, 1.4509704, 1.5892758, np.nan]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row, [0.1342734, 0.1257198, 0.0426454, np.nan]
    )
    np.testing.assert_almost_equal(slice_.scale_median_row, [3.0, 3.0, 3.0, np.nan])
    assert slice_.scale_median_row_margin == 3
    assert slice_.scale_median_column_margin is None
    assert slice_.scale_std_dev_column is None
    assert slice_.scale_std_err_column is None
    assert slice_.scale_median_column is None

    slice_ = Cube(CR.CA_X_MR_HS, transforms=transforms).partitions[1]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row, [1.3729408, 1.4601345, 1.4346546, np.nan]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row, [0.125546, 0.1265282, 0.0384517, np.nan]
    )
    np.testing.assert_almost_equal(slice_.scale_median_row, [3.0, 3.0, 4.0, np.nan])
    assert slice_.scale_median_row_margin == 3
    assert slice_.scale_median_column_margin is None
    assert slice_.scale_std_dev_column is None
    assert slice_.scale_std_err_column is None
    assert slice_.scale_median_column is None

    slice_ = Cube(CR.CA_X_MR_HS, transforms=transforms).partitions[2]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row, [np.nan, np.nan, np.nan, np.nan]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row, [np.nan, np.nan, np.nan, np.nan]
    )
    np.testing.assert_almost_equal(
        slice_.scale_median_row, [np.nan, np.nan, np.nan, np.nan]
    )
    assert slice_.scale_median_row_margin is None
    assert slice_.scale_median_column_margin is None
    assert slice_.scale_std_dev_column is None
    assert slice_.scale_std_err_column is None
    assert slice_.scale_median_column is None

    transforms = {
        "rows_dimension": {"prune": True},
        "columns_dimension": {"prune": True},
    }
    slice_ = Cube(CR.CA_X_MR_HS, transforms=transforms).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row, [1.4689114, 1.4509704, 1.5892758]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row, [0.1342734, 0.1257198, 0.0426454]
    )
    np.testing.assert_almost_equal(slice_.scale_median_row, [3, 3, 3])
    assert slice_.scale_median_row_margin == 3
    assert slice_.scale_median_column_margin is None
    assert slice_.scale_std_dev_column is None
    assert slice_.scale_std_err_column is None
    assert slice_.scale_median_column is None

    slice_ = Cube(CR.CA_X_MR_HS, transforms=transforms).partitions[1]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row, [1.3729408, 1.4601345, 1.4346546]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row, [0.125546, 0.1265282, 0.0384517]
    )
    np.testing.assert_almost_equal(slice_.scale_median_row, [3, 3, 4])
    assert slice_.scale_median_row_margin == 3
    assert slice_.scale_median_column_margin is None
    assert slice_.scale_std_dev_column is None
    assert slice_.scale_std_err_column is None
    assert slice_.scale_median_column is None

    slice_ = Cube(CR.CA_X_MR_HS, transforms=transforms).partitions[2]
    np.testing.assert_almost_equal(slice_.scale_std_dev_row, [])
    np.testing.assert_almost_equal(slice_.scale_std_err_row, [])
    np.testing.assert_almost_equal(slice_.scale_median_row, [])
    assert slice_.scale_median_row_margin is None
    assert slice_.scale_median_column_margin is None
    assert slice_.scale_std_dev_column is None
    assert slice_.scale_std_err_column is None
    assert slice_.scale_median_column is None


def test_cat_x_cat_pruning_and_hs():
    transforms = {
        "columns_dimension": {"insertions": {}},
        "rows_dimension": {"insertions": {}},
    }
    slice_ = Cube(CR.CAT_X_CAT_PRUNING_HS, transforms=transforms).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row, [1.2024596, 1.4649884, 1.5505837, np.nan, 0.9379386]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_column,
        [0.8506362, 0.9995499, 1.3697947, 0.6971257, np.nan, 0.8164966],
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row, [0.1994903, 0.2543296, 0.3564133, np.nan, 0.5440022]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_column,
        [0.1102738, 0.7064704, 0.4111442, 0.17486, np.nan, 0.4784233],
    )
    np.testing.assert_almost_equal(slice_.scale_median_row, [1, 1, 1, np.nan, 3])
    np.testing.assert_almost_equal(
        slice_.scale_median_column, [2, 1, 2, 2, np.nan, np.nan]
    )
    assert slice_.scale_median_row_margin == 1
    assert slice_.scale_median_column_margin == 2

    # Just H&S
    slice_ = Cube(CR.CAT_X_CAT_PRUNING_HS).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row,
        [1.2024596, 1.359933, 1.4649884, 1.5505837, np.nan, 0.9379386],
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_column,
        [0.8506362, 1.0412664, 0.9995499, 1.3697947, 0.6971257, np.nan, 0.8164966],
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row,
        [0.1994903, 0.163112, 0.2543296, 0.3564133, np.nan, 0.5440022],
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_column,
        [0.1102738, 0.1933713, 0.7064704, 0.4111442, 0.17486, np.nan, 0.4784233],
    )
    np.testing.assert_almost_equal(slice_.scale_median_row, [1, 1, 1, 1, np.nan, 3])
    np.testing.assert_almost_equal(
        slice_.scale_median_column, [2, 2, 1, 2, 2, np.nan, np.nan]
    )
    assert slice_.scale_median_row_margin == 1
    assert slice_.scale_median_column_margin == 2

    # Just pruning
    transforms = {
        "rows_dimension": {"prune": True},
        "columns_dimension": {"prune": True},
    }
    slice_ = Cube(CR.CAT_X_CAT_PRUNING_HS, transforms=transforms).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row, [1.2024596, 1.359933, 1.4649884, 1.5505837, 0.9379386]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_column,
        [0.8506362, 1.0412664, 0.9995499, 1.3697947, 0.6971257, 0.8164966],
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row, [0.1994903, 0.163112, 0.2543296, 0.3564133, 0.5440022]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_column,
        [0.1102738, 0.1933713, 0.7064704, 0.4111442, 0.17486, 0.4784233],
    )
    np.testing.assert_almost_equal(slice_.scale_median_row, [1, 1, 1, 1, 3])
    np.testing.assert_almost_equal(slice_.scale_median_column, [2, 2, 1, 2, 2, np.nan])
    assert slice_.scale_median_row_margin == 1
    assert slice_.scale_median_column_margin == 2

    # Pruning and H&S
    transforms = {
        "rows_dimension": {"insertions": {}, "prune": True},
        "columns_dimension": {"insertions": {}, "prune": True},
    }
    slice_ = Cube(CR.CAT_X_CAT_PRUNING_HS, transforms=transforms).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row, [1.2024596, 1.4649884, 1.5505837, 0.9379386]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_column,
        [0.8506362, 0.9995499, 1.3697947, 0.6971257, 0.8164966],
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row, [0.1994903, 0.2543296, 0.3564133, 0.5440022]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_column,
        [0.1102738, 0.7064704, 0.4111442, 0.17486, 0.4784233],
    )
    np.testing.assert_almost_equal(slice_.scale_median_row, [1, 1, 1, 3])
    np.testing.assert_almost_equal(slice_.scale_median_column, [2, 1, 2, 2, np.nan])
    assert slice_.scale_median_row_margin == 1
    assert slice_.scale_median_column_margin == 2


def test_cat_single_element_x_cat():
    slice_ = Cube(SM.CAT_SINGLE_ELEMENT_X_CAT).partitions[0]
    np.testing.assert_equal(slice_.scale_std_dev_row, [np.nan, np.nan, np.nan, np.nan])
    np.testing.assert_equal(slice_.scale_std_dev_column, [np.nan])
    np.testing.assert_equal(slice_.scale_std_err_row, [np.nan, np.nan, np.nan, np.nan])
    np.testing.assert_equal(slice_.scale_std_err_column, [np.nan])
    np.testing.assert_almost_equal(
        slice_.scale_median_row, [np.nan, np.nan, np.nan, np.nan]
    )
    np.testing.assert_almost_equal(slice_.scale_median_column, [np.nan])
    assert slice_.scale_median_row_margin is None
    assert slice_.scale_median_column_margin is None


def test_bivariate_cat():
    slice_ = Cube(CR.ECON_BLAME_X_IDEOLOGY_ROW_HS).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row,
        [0.7195463, 0.7196963, 0.9977753, 1.0608932, 1.0948414, 1.5740076],
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row,
        [0.0847993, 0.0533474, 0.0515249, 0.0718528, 0.104389, 0.2488725],
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_column,
        [0.94303101, 0.96775835, 1.16801487, 0.98177679, 1.88561808, 1.5987533],
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_column,
        [0.0558603, 0.0486317, 0.0447584, 0.063111, 0.7698004, 0.1938773],
    )
    np.testing.assert_almost_equal(slice_.scale_median_row, [2, 2, 2, 1, 1, 5])
    np.testing.assert_almost_equal(slice_.scale_median_column, [4, 3, 3, 3, 3.5, 4])
    assert slice_.scale_median_row_margin == 2
    assert slice_.scale_median_column_margin == 3


def test_cat_array_cat_dim_first():
    slice_ = Cube(CR.PETS_ARRAY_CAT_FIRST).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_column, [0.4967781, 0.4996192, 0.4937496]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_column, [0.0351443, 0.0347574, 0.0366944]
    )
    assert slice_.scale_median_column_margin == 1
    assert slice_.scale_std_dev_row is None
    assert slice_.scale_std_err_row is None
    assert slice_.scale_median_row_margin is None


def test_cat_array_subvar_dim_first():
    slice_ = Cube(CR.PETS_ARRAY_SUBVAR_FIRST).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row, [0.4967781, 0.4996192, 0.4937496]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row, [0.0351443, 0.0347574, 0.0366944]
    )
    np.testing.assert_almost_equal(slice_.scale_median_row, [1, 1, 2])
    assert slice_.scale_median_row_margin == 1
    assert slice_.scale_median_column_margin is None
    assert slice_.scale_std_dev_column is None
    assert slice_.scale_std_err_column is None
    assert slice_.scale_median_column is None


def test_cat_x_cat_arr_fruit_first():
    slice_ = Cube(CR.FRUIT_X_PETS_ARRAY).partitions[0]
    assert slice_.scale_std_dev_row is None
    assert slice_.scale_std_err_row is None
    assert slice_.scale_median_row is None
    assert slice_.scale_median_row_margin is None
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_column, [0.4995998, 0.4948717, 0.4995272]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_column, [0.09992, 0.093522, 0.1041586]
    )
    np.testing.assert_almost_equal(slice_.scale_median_column, [1, 1, 2])
    assert slice_.scale_median_column_margin == 1

    slice_ = Cube(CR.FRUIT_X_PETS_ARRAY).partitions[1]
    assert slice_.scale_std_dev_row is None
    assert slice_.scale_std_err_row is None
    assert slice_.scale_median_row is None
    assert slice_.scale_median_row_margin is None
    assert slice_.scale_median_column_margin == 1
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_column, [0.4913518, 0.4985185, 0.4971626]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_column, [0.0668645, 0.0691321, 0.0725186]
    )
    np.testing.assert_almost_equal(slice_.scale_median_column, [1, 2, 2])


def test_cat_x_cat_arr_subvars_first():
    slice_ = Cube(CR.FRUIT_X_PETS_ARRAY_SUBVARS_FIRST).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row, [0.4532462, 0.4898979, 0.4749589]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row, [0.067566, 0.0774597, 0.0839617]
    )
    np.testing.assert_almost_equal(slice_.scale_median_row, [2, 2, 2])
    assert slice_.scale_median_row_margin == 2
    assert slice_.scale_median_column_margin is None
    assert slice_.scale_std_dev_column is None
    assert slice_.scale_std_err_column is None
    assert slice_.scale_median_column is None

    slice_ = Cube(CR.FRUIT_X_PETS_ARRAY_SUBVARS_FIRST).partitions[1]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row, [0.4778846, 0.4582576, 0.4648295]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row, [0.0819565, 0.0724569, 0.0754053]
    )
    np.testing.assert_almost_equal(slice_.scale_median_row, [2, 2, 2])
    assert slice_.scale_median_column_margin is None
    assert slice_.scale_std_dev_column is None
    assert slice_.scale_std_err_column is None
    assert slice_.scale_median_column is None


def test_cat_x_cat_arr_pets_first():
    slice_ = Cube(CR.FRUIT_X_PETS_ARRAY_PETS_FIRST).partitions[0]
    np.testing.assert_almost_equal(slice_.scale_std_dev_row, [0.4995998, 0.4913518])
    np.testing.assert_almost_equal(slice_.scale_std_dev_column, [0.4532462, 0.4778846])
    np.testing.assert_almost_equal(slice_.scale_std_err_row, [0.09992, 0.0668645])
    np.testing.assert_almost_equal(slice_.scale_std_err_column, [0.067566, 0.0819565])
    np.testing.assert_almost_equal(slice_.scale_median_row, [1, 1])
    np.testing.assert_almost_equal(slice_.scale_median_column, [2, 2])
    assert slice_.scale_median_row_margin == 1
    assert slice_.scale_median_column_margin == 2

    slice_ = Cube(CR.FRUIT_X_PETS_ARRAY_PETS_FIRST).partitions[1]
    np.testing.assert_almost_equal(slice_.scale_std_dev_row, [0.4948717, 0.4985185])
    np.testing.assert_almost_equal(slice_.scale_std_dev_column, [0.4898979, 0.4582576])
    np.testing.assert_almost_equal(slice_.scale_std_err_row, [0.093522, 0.0691321])
    np.testing.assert_almost_equal(slice_.scale_std_err_column, [0.0774597, 0.0724569])
    np.testing.assert_almost_equal(slice_.scale_median_row, [1, 2])
    np.testing.assert_almost_equal(slice_.scale_median_column, [2, 2])
    assert slice_.scale_median_row_margin == 1.5
    assert slice_.scale_median_column_margin == 2

    slice_ = Cube(CR.FRUIT_X_PETS_ARRAY_PETS_FIRST).partitions[2]
    np.testing.assert_almost_equal(slice_.scale_std_dev_row, [0.4995272, 0.4971626])
    np.testing.assert_almost_equal(slice_.scale_std_dev_column, [0.4749589, 0.4648295])
    np.testing.assert_almost_equal(slice_.scale_std_err_row, [0.1041586, 0.0725186])
    np.testing.assert_almost_equal(slice_.scale_std_err_column, [0.0839617, 0.0754053])
    np.testing.assert_almost_equal(slice_.scale_median_row, [2, 2])
    np.testing.assert_almost_equal(slice_.scale_median_column, [2, 2])
    assert slice_.scale_median_row_margin == 2
    assert slice_.scale_median_column_margin == 2


def test_with_null_values():
    slice_ = Cube(CR.SCALE_WITH_NULL_VALUES).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_std_dev_row, [0.911757, 0.9089713, 0.9419575]
    )
    np.testing.assert_almost_equal(
        slice_.scale_std_err_row, [0.0200883, 0.0178672, 0.016882]
    )
    np.testing.assert_almost_equal(slice_.scale_median_row, [1, 1, 1])
    assert slice_.scale_median_row_margin == 1
    assert slice_.scale_median_column_margin is None
    assert slice_.scale_std_dev_column is None
    assert slice_.scale_std_err_column is None
    assert slice_.scale_median_column is None



# encoding: utf-8

"""Unit test suite for `cr.cube.cubepart` module."""

from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np
import pytest

from cr.cube.cube import Cube
from cr.cube.cubepart import CubePartition, _Slice, _Strand, _Nub
from cr.cube.dimension import Dimension
from cr.cube.enums import DIMENSION_TYPE as DT
from cr.cube.matrix import TransformedMatrix, _VectorAfterHiding
from cr.cube.noa.smoothing import SingleSidedMovingAvgSmoother
from cr.cube.stripe import _BaseStripeRow, TransformedStripe

from ..unitutil import class_mock, instance_mock, property_mock


class DescribeCubePartition(object):
    """Unit test suite for `cr.cube.cubepart.CubePartition` object."""

    def it_constructs_a_slice_with_its_factory_for_a_2D_cube(self, request, cube_):
        cube_.ndim = 2
        _Slice_ = class_mock(request, "cr.cube.cubepart._Slice")

        slice_ = CubePartition.factory(cube_, 42, {"trans": "forms"}, 1000, False, 10)

        _Slice_.assert_called_once_with(cube_, 42, {"trans": "forms"}, 1000, 10)
        assert slice_ is _Slice_.return_value

    def but_it_constructs_a_strand_for_a_2D_cube_when_ca_as_0th(self, request, cube_):
        cube_.ndim = 2
        _Strand_ = class_mock(request, "cr.cube.cubepart._Strand")

        strand = CubePartition.factory(cube_, 0, ca_as_0th=True)

        _Strand_.assert_called_once_with(cube_, None, None, True, 0, 0)
        assert strand is _Strand_.return_value

    def and_it_constructs_a_strand_for_a_1D_cube(self, request, cube_):
        cube_.ndim = 1
        _Strand_ = class_mock(request, "cr.cube.cubepart._Strand")

        strand = CubePartition.factory(cube_, 42, {"trans": "forms"}, 1000, False, 10)

        _Strand_.assert_called_once_with(cube_, {"trans": "forms"}, 1000, False, 42, 10)
        assert strand is _Strand_.return_value

    def and_it_constructs_a_nub_for_a_0D_cube(self, request, cube_):
        cube_.ndim = 0
        _Nub_ = class_mock(request, "cr.cube.cubepart._Nub")

        nub = CubePartition.factory(cube_)

        _Nub_.assert_called_once_with(cube_)
        assert nub is _Nub_.return_value

    def it_knows_the_index_of_its_cube_in_the_cube_set(self, cube_):
        cube_.cube_index = 42
        cube_partition = CubePartition(cube_)

        cube_index = cube_partition.cube_index

        assert cube_index == 42

    def it_knows_if_cube_is_mr_aug(self):
        # --- default of False is overridden by subclasses when appropriate ---
        assert CubePartition(None).cube_is_mr_aug is False

    def it_can_evaluate_a_measure_expression(self, request):
        single_sided_moving_avg_smoother_ = instance_mock(
            request, SingleSidedMovingAvgSmoother, values=[[0.1, 0.2], [0.3, 0.4]]
        )
        SingleSidedMovingAvgSmoother_ = class_mock(
            request,
            "cr.cube.cubepart.SingleSidedMovingAvgSmoother",
            return_value=single_sided_moving_avg_smoother_,
        )
        measure_expr = {
            "function": "one_sided_moving_avg",
            "base_measure": "col_percent",
            "window": 2,
        }
        cube_partition = CubePartition(None, None)

        values = cube_partition.evaluate(measure_expr)

        SingleSidedMovingAvgSmoother_.assert_called_once_with(
            cube_partition, measure_expr
        )
        assert values == [[0.1, 0.2], [0.3, 0.4]]

    @pytest.mark.parametrize(
        "dims, expected_value",
        (
            ((DT.CAT, DT.MR, DT.MR_CAT), DT.CAT),
            ((DT.MR, DT.MR_CAT), DT.MR),
            ((DT.MR_CAT,), DT.MR_CAT),
        ),
    )
    def it_knows_cube_row_dimension_type(self, request, cube_, dims, expected_value):
        all_dimensions_ = tuple(
            instance_mock(request, Dimension, name="dim-%d" % idx, dimension_type=dt)
            for idx, dt in enumerate(dims)
        )
        cube_.dimensions = all_dimensions_
        cube_partition = CubePartition(cube_)

        cube_row_dimension_type = cube_partition.cube_row_dimension_type

        assert cube_row_dimension_type == expected_value

    def it_knows_the_primary_alpha_value_to_help(self, _alpha_values_prop_):
        """alpha is the primary confidence-interval threshold specified by the user."""
        _alpha_values_prop_.return_value = (0.042, 0.084)
        assert CubePartition(None)._alpha == 0.042

    @pytest.mark.parametrize(
        "alpha_values, expected_value", (((0.042, 0.084), 0.084), ((0.042, None), None))
    )
    def it_knows_the_secondary_alpha_value_to_help(
        self, _alpha_values_prop_, alpha_values, expected_value
    ):
        _alpha_values_prop_.return_value = alpha_values
        assert CubePartition(None)._alpha_alt == expected_value

    @pytest.mark.parametrize(
        "pw_indices_dict, expected_value",
        (
            # --- default value is .05 ---
            ({}, (0.05, None)),
            ({"alpha": {}}, (0.05, None)),
            ({"alpha": []}, (0.05, None)),
            # --- scalar (float) value sets alpha and sets alpha_alt to None ---
            ({"alpha": 0.025}, (0.025, None)),
            # --- single (float) value sets alpha and sets alpha_alt to None ---
            ({"alpha": [0.03]}, (0.03, None)),
            # --- two values sets alpha to lesser and alpha_alt to greater ---
            ({"alpha": [0.07, 0.03]}, (0.03, 0.07)),
            # --- third and later values are ignored ---
            ({"alpha": (0.07, 0.03, "foobar")}, (0.03, 0.07)),
        ),
    )
    def it_interprets_the_provided_alpha_values_to_help(
        self, pw_indices_dict, expected_value
    ):
        cube_partition = CubePartition(None, {"pairwise_indices": pw_indices_dict})

        alpha_values = cube_partition._alpha_values

        assert alpha_values == expected_value

    @pytest.mark.parametrize(
        "pw_indices_dict, exception_type, expected_message",
        (
            # --- type errors ---
            (
                {"alpha": {"al": "pha"}},
                TypeError,
                "transforms.pairwise_indices.alpha, when defined, must be a list of 1 "
                "or 2 float values between 0.0 and 1.0 exclusive. Got %s"
                % repr({"al": "pha"}),
            ),
            (
                {"alpha": "0.05"},
                TypeError,
                "transforms.pairwise_indices.alpha, when defined, must be a list of 1 "
                "or 2 float values between 0.0 and 1.0 exclusive. Got %s"
                % repr("0.05"),
            ),
            # --- scalar out-of-range errors ---
            (
                {"alpha": -0.1},
                ValueError,
                "alpha value, when provided, must be between 0.0 and 1.0 exclusive. "
                "Got %s" % repr(-0.1),
            ),
            (
                {"alpha": 1.0},
                ValueError,
                "alpha value, when provided, must be between 0.0 and 1.0 exclusive. "
                "Got %s" % repr(1.0),
            ),
            # --- sequence value errors ---
            (
                {"alpha": [0.01, ".05"]},
                ValueError,
                "transforms.pairwise_indices.alpha must be a list of 1 or 2 float "
                "values between 0.0 and 1.0 exclusive. Got %s" % repr([0.01, ".05"]),
            ),
        ),
    )
    def but_it_raises_on_invalid_alpha_values(
        self, pw_indices_dict, exception_type, expected_message
    ):
        cube_partition = CubePartition(None, {"pairwise_indices": pw_indices_dict})

        with pytest.raises(exception_type) as e:
            cube_partition._alpha_values

        assert str(e.value) == expected_message

    @pytest.mark.parametrize(
        "pw_indices_dict, expected_value",
        (
            # --- default value is True ---
            ({}, True),
            ({"only_larger": "foobar"}, True),
            ({"only_larger": False}, False),
        ),
    )
    def it_knows_the_only_larger_flag_state_to_help(
        self, _transforms_dict_prop_, pw_indices_dict, expected_value
    ):
        _transforms_dict_prop_.return_value = {"pairwise_indices": pw_indices_dict}
        assert CubePartition(None)._only_larger == expected_value

    @pytest.mark.parametrize(
        "transforms, expected_value",
        ((None, {}), ({"trans": "forms"}, {"trans": "forms"})),
    )
    def it_provides_the_transforms_dict_to_help(self, transforms, expected_value):
        """Handles defaulting of transforms arg."""
        assert CubePartition(None, transforms)._transforms_dict == expected_value

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _alpha_values_prop_(self, request):
        return property_mock(request, CubePartition, "_alpha_values")

    @pytest.fixture
    def cube_(self, request):
        return instance_mock(request, Cube)

    @pytest.fixture
    def _transforms_dict_prop_(self, request):
        return property_mock(request, CubePartition, "_transforms_dict")


class Describe_Slice(object):
    """Unit test suite for `cr.cube.cubepart._Slice` object."""

    @pytest.mark.parametrize(
        ("shape", "expected_value"),
        (((4, 2), False), ((4, 0), True), ((0, 2), True), ((0, 0), True)),
    )
    def it_knows_whether_it_is_empty(self, shape, expected_value, shape_prop_):
        shape_prop_.return_value = shape
        slice_ = _Slice(None, None, None, None, None)

        is_empty = slice_.is_empty

        assert is_empty is expected_value

    def it_provides_the_secondary_pairwise_indices(
        self, _alpha_alt_prop_, _only_larger_prop_, PairwiseSignificance_
    ):
        PairwiseSignificance_.pairwise_indices.return_value = [[(0,), (1,)], [(), ()]]
        _alpha_alt_prop_.return_value = 0.42
        _only_larger_prop_.return_value = False
        slice_ = _Slice(None, None, None, None, None)

        pairwise_indices_alt = slice_.pairwise_indices_alt

        PairwiseSignificance_.pairwise_indices.assert_called_once_with(
            slice_, 0.42, False
        )
        assert pairwise_indices_alt == [[(0,), (1,)], [(), ()]]

    def but_it_returns_None_when_no_secondary_alpha_specified(self, _alpha_alt_prop_):
        _alpha_alt_prop_.return_value = None
        assert _Slice(None, None, None, None, None).pairwise_indices_alt is None

    def it_knows_the_population_fraction(self, cube_):
        cube_.population_fraction = 0.5
        slice_ = _Slice(cube_, None, None, None, None)

        population_fraction = slice_.population_fraction

        assert population_fraction == 0.5

    def it_knows_the_row_proportions(self, request, _matrix_prop_, matrix_):
        _matrix_prop_.return_value = matrix_
        matrix_.rows = (
            instance_mock(request, _VectorAfterHiding, proportions=(0.1, 0.2, 0.3)),
            instance_mock(request, _VectorAfterHiding, proportions=(0.4, 0.5, 0.6)),
        )
        slice_ = _Slice(None, None, None, None, None)

        row_proportions = slice_.row_proportions

        np.testing.assert_almost_equal(
            row_proportions, [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]]
        )

    def it_knows_the_rows_margin(self, request, _matrix_prop_, matrix_):
        _matrix_prop_.return_value = matrix_
        matrix_.rows = (
            instance_mock(request, _VectorAfterHiding, margin=(1, 2)),
            instance_mock(request, _VectorAfterHiding, margin=(3, 4)),
        )
        slice_ = _Slice(None, None, None, None, None)

        rows_margin = slice_.rows_margin

        np.testing.assert_almost_equal(rows_margin, [[1, 2], [3, 4]])

    def it_provides_the_secondary_scale_mean_pairwise_indices(
        self, _alpha_alt_prop_, _only_larger_prop_, PairwiseSignificance_
    ):
        PairwiseSignificance_.scale_mean_pairwise_indices.return_value = (
            (2,),
            (0,),
            (),
        )
        _alpha_alt_prop_.return_value = 0.42
        _only_larger_prop_.return_value = True
        slice_ = _Slice(None, None, None, None, None)

        scale_mean_pairwise_indices_alt = slice_.scale_mean_pairwise_indices_alt

        PairwiseSignificance_.scale_mean_pairwise_indices.assert_called_once_with(
            slice_, 0.42, True
        )
        assert scale_mean_pairwise_indices_alt == ((2,), (0,), ())

    def but_scale_mean_pairwise_indices_alt_is_None_when_no_secondary_alpha_specified(
        self, _alpha_alt_prop_
    ):
        _alpha_alt_prop_.return_value = None
        slice_ = _Slice(None, None, None, None, None)

        assert slice_.scale_mean_pairwise_indices_alt is None

    def but_it_raises_an_exception_when_function_is_not_available(self):
        slice_ = _Slice(None, None, None, None, None)

        with pytest.raises(NotImplementedError) as err:
            slice_.evaluate({"function": "F"})

        assert str(err.value) == "Function F is not available."

    @pytest.mark.parametrize(
        "dimensions_dicts, expected_value",
        (
            ([{"references": {}}], ()),
            (
                [
                    {"references": {}},
                    {
                        "references": {
                            "selected_categories": [{"name": "Foo"}, {"name": "Bar"}]
                        }
                    },
                ],
                ("Foo", "Bar"),
            ),
            (
                [
                    {"references": {}},
                    {
                        "references": {
                            "selected_categories": [{"name": "Foo"}, {"id": "Bar"}]
                        }
                    },
                ],
                ("Foo",),
            ),
        ),
    )
    def it_knows_its_selected_category_labels(
        self, _dimensions_prop_, dimensions_dicts, expected_value
    ):
        _dimensions_prop_.return_value = [
            Dimension(dim_dict, None, None) for dim_dict in dimensions_dicts
        ]
        slice_ = _Slice(None, None, None, None, None)

        assert slice_.selected_category_labels == expected_value

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _alpha_alt_prop_(self, request):
        return property_mock(request, _Slice, "_alpha_alt")

    @pytest.fixture
    def cube_(self, request):
        return instance_mock(request, Cube)

    @pytest.fixture
    def _dimensions_prop_(self, request):
        return property_mock(request, _Slice, "_dimensions")

    @pytest.fixture
    def matrix_(self, request):
        return instance_mock(request, TransformedMatrix)

    @pytest.fixture
    def _matrix_prop_(self, request):
        return property_mock(request, _Slice, "_matrix")

    @pytest.fixture
    def _only_larger_prop_(self, request):
        return property_mock(request, _Slice, "_only_larger")

    @pytest.fixture
    def PairwiseSignificance_(self, request):
        return class_mock(request, "cr.cube.cubepart.PairwiseSignificance")

    @pytest.fixture
    def shape_prop_(self, request):
        return property_mock(request, _Slice, "shape")


class Describe_Strand(object):
    """Unit test suite for `cr.cube.cubepart._Strand` object."""

    def it_knows_which_of_its_rows_are_inserted(self, request, _stripe_prop_, stripe_):
        stripe_.rows = tuple(
            instance_mock(request, _BaseStripeRow, is_inserted=bool(i % 2))
            for i in range(7)
        )
        _stripe_prop_.return_value = stripe_
        strand = _Strand(None, None, None, None, None, None)

        inserted_row_idxs = strand.inserted_row_idxs

        assert inserted_row_idxs == (1, 3, 5)

    @pytest.mark.parametrize(("shape", "expected_value"), (((1,), False), ((0,), True)))
    def it_knows_whether_it_is_empty(self, shape, expected_value, shape_prop_):
        shape_prop_.return_value = shape
        strand = _Strand(None, None, None, None, None, None)

        is_empty = strand.is_empty

        assert is_empty is expected_value

    def it_knows_its_title(self, cube_):
        cube_.title = "Unmarried"
        strand_ = _Strand(cube_, None, None, None, None, None)

        title = strand_.title

        assert title == "Unmarried"

    def it_constructs_its_underlying_stripe_to_help(
        self, request, cube_, _rows_dimension_prop_, dimension_, stripe_
    ):
        TransformedStripe_ = class_mock(request, "cr.cube.cubepart.TransformedStripe")
        TransformedStripe_.stripe.return_value = stripe_
        _rows_dimension_prop_.return_value = dimension_
        strand = _Strand(cube_, None, None, False, 42, None)

        stripe = strand._stripe

        TransformedStripe_.stripe.assert_called_once_with(cube_, dimension_, False, 42)
        assert stripe is stripe_

    def it_knows_the_population_fraction(self, cube_):
        cube_.population_fraction = 0.5
        strand_ = _Strand(cube_, None, None, None, None, None)

        population_fraction = strand_.population_fraction

        assert population_fraction == 0.5

    def it_knows_its_selected_categories_labels(self, _dimensions_prop_):
        _dimensions_prop_.return_value = [Dimension({"references": {}}, None, None)]
        strand_ = _Strand(None, None, None, None, None, None)

        assert strand_.selected_category_labels == ()

    # fixture components ---------------------------------------------

    @pytest.fixture
    def cube_(self, request):
        return instance_mock(request, Cube)

    @pytest.fixture
    def dimension_(self, request):
        return instance_mock(request, Dimension)

    @pytest.fixture
    def _dimensions_prop_(self, request):
        return property_mock(request, _Strand, "_dimensions")

    @pytest.fixture
    def _rows_dimension_prop_(self, request):
        return property_mock(request, _Strand, "_rows_dimension")

    @pytest.fixture
    def shape_prop_(self, request):
        return property_mock(request, _Strand, "shape")

    @pytest.fixture
    def stripe_(self, request):
        return instance_mock(request, TransformedStripe)

    @pytest.fixture
    def _stripe_prop_(self, request):
        return property_mock(request, _Strand, "_stripe")


class Describe_Nub(object):
    """Unit test suite for `cr.cube.cubepart._Nub` object."""

    @pytest.mark.parametrize(
        "unweighted_count, expected_value", ((None, True), (45.4, False))
    )
    def it_knows_when_it_is_empty(self, request, unweighted_count, expected_value):
        property_mock(request, _Nub, "unweighted_count", return_value=unweighted_count)
        nub_ = _Nub(None)

        is_empty = nub_.is_empty

        assert is_empty == expected_value

    def it_knows_its_cube_is_never_mr_aug(self):
        assert _Nub(None).cube_is_mr_aug is False

    def it_knows_its_selected_categories_labels(self):
        nub_ = _Nub(None)

        assert nub_.selected_category_labels == ()



# encoding: utf-8

"""Integration-test suite for smoothing feature."""

import numpy as np
import pytest

from cr.cube.cube import Cube

from ..fixtures import CR
from ..util import load_python_expression


class DescribeSliceSmoothing(object):
    """Integration-test suite for _Slice.evaluate() method."""

    @pytest.mark.parametrize(
        "fixture, window, expectation",
        (
            (CR.CAT_X_CAT_DATE_WGTD, 4, "cat-x-cat-date-wgtd-smoothed-col-idx-w4"),
            (CR.CAT_X_MR_X_CAT_DATE, 3, "cat-x-mr-x-cat-date-smoothed-col-idx-w3"),
            (
                CR.CA_SUBVAR_X_CA_CAT_X_CAT_DATE,
                3,
                "ca-subvar-x-ca-cat-cat-date-smoothed-col-idx-w3",
            ),
            (CR.CAT_HS_X_CAT_DATE, 3, "cat-hs-x-cat-date-smoothed-col-idx-w3"),
            (CR.MR_X_CAT_DATE, 3, "mr-x-cat-date-smoothed-col-idx-w3"),
            (CR.NUMERIC_X_CAT_DATE, 3, "numeric-x-cat-date-smoothed-col-idx-w3"),
            (CR.TXT_X_CAT_DATE, 3, "txt-x-cat-date-smoothed-col-idx-w3"),
            (CR.DATETIME_X_CAT_DATE, 3, "datetime-x-cat-date-smoothed-col-idx-w3"),
        ),
    )
    def it_provides_smoothed_col_index_for_compatible_cubes(
        self, fixture, window, expectation
    ):
        slice_ = Cube(fixture).partitions[0]
        col_index = slice_.evaluate(
            {
                "function": "one_sided_moving_avg",
                "base_measure": "col_index",
                "window": window,
            }
        )
        np.testing.assert_array_almost_equal(
            col_index, load_python_expression(expectation)
        )

    @pytest.mark.parametrize(
        "fixture, window, expectation",
        (
            (CR.CAT_X_CAT_DATE_WGTD, 4, "cat-x-cat-date-wgtd-smoothed-col-pct-w4"),
            (CR.CAT_X_MR_X_CAT_DATE, 3, "cat-x-mr-x-cat-date-smoothed-col-pct-w3"),
            (
                CR.CA_SUBVAR_X_CA_CAT_X_CAT_DATE,
                3,
                "ca-subvar-x-ca-cat-cat-date-smoothed-col-pct-w3",
            ),
            (CR.CAT_HS_X_CAT_DATE, 3, "cat-hs-x-cat-date-smoothed-col-pct-w3"),
            (CR.MR_X_CAT_DATE, 3, "mr-x-cat-date-smoothed-col-pct-w3"),
            (CR.NUMERIC_X_CAT_DATE, 3, "numeric-x-cat-date-smoothed-col-pct-w3"),
            (CR.TXT_X_CAT_DATE, 3, "txt-x-cat-date-smoothed-col-pct-w3"),
            (CR.DATETIME_X_CAT_DATE, 3, "datetime-x-cat-date-smoothed-col-pct-w3"),
        ),
    )
    def it_provides_smoothed_col_percent_for_compatible_cubes(
        self, fixture, window, expectation
    ):
        slice_ = Cube(fixture).partitions[0]
        col_percent = slice_.evaluate(
            {
                "function": "one_sided_moving_avg",
                "base_measure": "col_percent",
                "window": window,
            }
        )
        np.testing.assert_array_almost_equal(
            col_percent, load_python_expression(expectation)
        )

    @pytest.mark.parametrize(
        "fixture, expectation",
        (
            (CR.CAT_X_CAT_DATE, "cat-x-cat-date-smoothed-scale-means-w2"),
            (CR.CAT_X_CAT_DATE_WGTD, "cat-x-cat-date-smoothed-scale-means-w2"),
            (
                CR.CA_SUBVAR_X_CA_CAT_X_CAT_DATE,
                "ca-subvar-ca-cat-x-cat-date-scale-means-w2",
            ),
        ),
    )
    def it_provides_smoothed_scale_means_for_compatible_cubes(
        self, fixture, expectation
    ):

        slice_ = Cube(fixture).partitions[0]
        # --- window not expressed get the default value : 2
        scale_mean = slice_.evaluate(
            {
                "function": "one_sided_moving_avg",
                "base_measure": "scale_mean",
                "window": 2,
            }
        )
        np.testing.assert_array_almost_equal(
            scale_mean, load_python_expression(expectation)
        )

    @pytest.mark.parametrize(
        "fixture",
        (CR.CAT_X_MR, CR.MR_X_MR, CR.MR_X_CA_CAT_X_CA_SUBVAR, CR.CAT_DATE_X_CAT),
    )
    def it_warns_and_does_not_smooth_when_dimension_is_not_smoothable(self, fixture):
        slice_ = Cube(fixture).partitions[0]
        base_values = slice_.column_percentages
        expected_warning_regex = (
            r"No smoothing performed. Column dimension must be a categorical date."
        )

        with pytest.warns(UserWarning, match=expected_warning_regex):
            smoothed_values = slice_.evaluate(
                {
                    "function": "one_sided_moving_avg",
                    "base_measure": "col_percent",
                    "window": 3,
                }
            )

        np.testing.assert_array_almost_equal(smoothed_values, base_values)

    @pytest.mark.parametrize(
        "fixture, base_measure, prop_name, periods, window",
        (
            (CR.CAT_X_CAT_DATE, "col_percent", "column_percentages", 4, 1),
            (CR.CAT_X_CAT_DATE, "col_index", "column_index", 4, 1),
            (CR.CAT_X_CAT, "col_percent", "column_percentages", 2, 3),
        ),
    )
    def it_warns_and_does_not_smooth_when_window_is_invalid(
        self, fixture, base_measure, prop_name, periods, window
    ):
        slice_ = Cube(fixture).partitions[0]
        base_values = getattr(slice_, prop_name)
        expected_warning_regex = (
            r"No smoothing performed. Smoothing window must be between 2 and the "
            r"number of periods \(%d\), got %d" % (periods, window)
        )

        with pytest.warns(UserWarning, match=expected_warning_regex):
            smoothed_values = slice_.evaluate(
                {
                    "function": "one_sided_moving_avg",
                    "base_measure": base_measure,
                    "window": window,
                }
            )

        np.testing.assert_array_almost_equal(smoothed_values, base_values)


class DescribeStrandMeansSmoothing(object):
    """Integration-test suite for _Strand.evaluate() method."""

    def it_provides_smoothed_means_cat_date(self):
        strand_ = Cube(CR.CAT_DATE_MEAN).partitions[0]
        means = strand_.evaluate(
            {"function": "one_sided_moving_avg", "base_measure": "mean", "window": 3}
        )
        np.testing.assert_array_almost_equal(
            means, [np.nan, np.nan, 2.65670765025029, 2.5774816240050358]
        )

    def it_does_not_smooth_means_mr_mean_filt_wgtd(self):
        strand_ = Cube(CR.MR_MEAN_FILT_WGTD).partitions[0]
        expected_warning_regex = (
            r"No smoothing performed. Column dimension must be a categorical date."
        )

        with pytest.warns(UserWarning, match=expected_warning_regex):
            means = strand_.evaluate(
                {
                    "function": "one_sided_moving_avg",
                    "base_measure": "mean",
                    "window": 3,
                }
            )
        np.testing.assert_array_almost_equal(
            means, [3.724051, 2.578429, 2.218593, 1.865335]
        )
